__all__ = ["MAGNITUDE_BASE", "STDFLUX", "calculate_magnitude", "get_vega_spectrum",
           "Bandpass", "UBVTabulated", "UBVParametric", "get_ubv_bandpasses",
           "get_zero_flux", "flux_to_mag", "mag_to_flux", "get_ubv_bandpass",
           "get_ubv_bandpasses_dict", "calc_mag"]


import numpy as np
import collections
from scipy.interpolate import interp1d
import copy
import a99
import f311.filetypes as ft
# from .. import physics as ph

MAGNITUDE_BASE = 100. ** (1. / 5)  # approx. 2.512
_REF_NUM_POINTS = 5000   # number of evaluation points over entire band range


_ubv_bandpasses = None
_ubv_bandpasses_dict = None
def get_ubv_bandpasses():
    """Returns list with UBVRI... Bandpass objects"""
    global _ubv_bandpasses, _ubv_bandpasses_dict
    if _ubv_bandpasses is None:
        _ubv_bandpasses = []
        _ubv_bandpasses_dict = collections.OrderedDict()
        for name in UBVParametric.X0_FWHM.keys():
            bp = UBVTabulated(name) if name in "UBVRI" else UBVParametric(name)
            _ubv_bandpasses.append(bp)
            _ubv_bandpasses_dict[name] = bp
    return _ubv_bandpasses


def get_ubv_bandpasses_dict():
    """Returns dictionary with band name, i.e., U/B/V/etc. as dict"""
    get_ubv_bandpasses()  # just to assure the dict is assembled
    return _ubv_bandpasses_dict


def get_ubv_bandpass(bp):
    """Returns a BandPass object for given band name

    Args:
        bp: band name or BandPass object

    Returns:
        f311.physics.BandPass: BandPass object. If `bp` is already a BandPass object, returns `bp`
    """
    if isinstance(bp, Bandpass):
        return bp
    get_ubv_bandpasses()  # just to assure the dict is assembled
    return _ubv_bandpasses_dict[bp]


def calc_mag(sp, bp, system="stdflux", zero_point=0., flag_force_band_range=False):
    """
    Calculates magnitude from a Spectrum object. See calculate_magnitude for arguments documentation

    Returns: (float) calculated magnitude

    Example:

    >>> # The following example prints a table showing the calculated magnitude for the
    >>> # Vega spectrum in several systems
    >>> import f311.physics as ph
    >>> import tabulate
    >>> systems = ["stdflux", "ab", "vega"]
    >>> bands = "UBVRIJHK"
    >>> sp = ph.get_vega_spectrum()
    >>> rows = [([band]+[ph.calc_mag(sp, band, system) for system in systems]) for band in bands]
    >>> print(tabulate.tabulate(rows, ["band"]+systems))
    band        stdflux          ab    vega
    ------  -----------  ----------  ------
    U        0.00572505   0.761594       -0
    B        0.0696287   -0.10383        -0
    V        0.0218067    0.0191189      -0
    R        0.0359559    0.214645       -0
    I        0.0661095    0.449825       -0
    J       -0.0150993    0.874666       -0
    H        0.0315447    1.34805        -0
    K        0.0246046    1.85948        -0
    """

    return calculate_magnitude(sp, bp, system, zero_point, flag_force_band_range)["cmag"]


def calculate_magnitude(sp, bp, system="stdflux", zero_point=0., flag_force_band_range=False):
    """
    Calculates magnitude from a Spectrum object.

    Args:
        sp: `f311.filetypes.Spectrum` instance. **flux unit** must be ``erg/cm**2/s/Hz`` aka "Fnu"
        bp: Bandpass object, or string in "UBVRIYJHKLMNQ". How it works:

            - If string in "UBVRI", creates a UBVTabular Bandpass object;
            - If string otherwise, creates a UBVParametric object.

        system: reference magnitude system. Possible values:

            - ``"stdflux"``: literature reference values for bands U,B,V,R,I,J,H,K only
            - ``"vega"``: uses the Vega star spectrum as a reference
            - ``"ab"``: AB[solute] magnitude system

        zero_point: subtracts this value from the calculated magnitude to implement some desired
                    correction.

        flag_force_band_range: this flag has effect when the spectrum does not completely overlap
            the bandpass filter. How it works:

            - ``False`` (default): restricts the weighted mean flux calculation to the overlap
              range between the spectrum and the filter
            - ``True``: zero-fill the spectrum to overlap the filter range completely

    Returns:
        dict: dictionary with the following keys:

            - ``"calc_l0"``: lower edge of wavelength interval considered,
            - ``"calc_lf"``: upper edge of wavelength interval considered,
            - ``"filtered_sp"``: filtered spectrum
            - ``"filtered_sp_area"``: area of filtered spectrum within [calc_l0, calc_lf]
            - ``"weighted_mean_flux"``: weighted mean of flux; weights are band values
            - ``"zero_flux"``: flux value for which magnitude is zero
            - ``"cmag"``: calculated magnitude

    Example:

    >>> from f311 import physics as ph
    >>> ph.calculate_magnitude(ph.get_vega_spectrum(), "U", "stdflux", 0, False)

    """

    bp = get_ubv_bandpass(bp)

    # # Determines areas
    filtered_sp = sp * bp
    if flag_force_band_range:
        calc_l0, calc_lf = bp.l0, bp.lf
    else:
        calc_l0, calc_lf = max(bp.l0, sp.l0), min(bp.lf, sp.lf)
    band_area = bp.area(calc_l0, calc_lf)
    # Note: discarding zeroes before integrating would make very little difference in time:
    #       ~47.5 against ~49.6 for the Vega spectrum with 8827 points
    filtered_sp_area = np.trapz(filtered_sp.y, filtered_sp.x)

    zero_flux = get_zero_flux(bp, system)

    weighted_mean_flux = filtered_sp_area/band_area
    cmag = -2.5 * np.log10(weighted_mean_flux / zero_flux) - zero_point

    ret = collections.OrderedDict((
    ("bp", bp),
    ("calc_l0", calc_l0),
    ("calc_lf", calc_lf),
    ("filtered_sp", filtered_sp),
    ("filtered_sp_area", filtered_sp_area),
    ("weighted_mean_flux", weighted_mean_flux),
    ("zero_flux", zero_flux),
    ("cmag", cmag),
    ))
    # if system == "vega":
    #     ret["filtered_vega_sp"] = filtered_vega_sp
    return ret




def flux_to_mag(flux, bp, system="stdflux", zero_point=0.):
    """
    Calculates magnitude for a single scalar flux value

    Args:
        flux: float (erg/cm**2/s/Hz), of Spectrum instance
        bp, system, zero_point: see calculate_magnitude()

    Returns: float
    """
    from f311 import physics as ph

    if isinstance(flux, ft.Spectrum):
        if flux.yunit != ph.fnu:
            raise ValueError("Spectrum y-unit must be '{}', not '{}'".format(ph.fnu, flux.yunit))
        flux = flux.y
    zero_flux = get_zero_flux(bp, system)
    return -2.5 * np.log10(flux / zero_flux) - zero_point


def get_zero_flux(bp, system="stdflux"):
    """
    Returns flux (erg/cm**2/s/Hz) for which magnitude is zero

    Args:
        bp: Bandpass object, or band name, e.g., "U"
        system: {"stdflux", "ab", "vega"} reference magnitude system

    For more on arguments, see calculate_magnitude()

    Returns: float
    """

    bp = __get_ubv_bandpass(bp)

    if system == "stdflux":
        if bp.name not in STDFLUX:
            raise ValueError("No standard flux for band '{}'".format(bp.name))
        zero_flux = STDFLUX[bp.name]
    elif system == "ab":
        zero_flux = 3631e-23
    elif system == "vega":
        vega_sp = __get_vega_spectrum()
        filtered_vega_sp = vega_sp * bp
        zero_flux = np.trapz(filtered_vega_sp.y, filtered_vega_sp.x) / bp.area(bp.l0, bp.lf)
    else:
        raise ValueError("Invalid reference magnitude system: '{0!s}'".format(system))
    return zero_flux


def mag_to_flux(mag, bp, system="stdflux"):
    """Inverse of calc_mag()"""
    zero_flux = get_zero_flux(bp, system)
    return zero_flux*10**(-.4*mag)


__vega_spectrum = None
def __get_vega_spectrum():
    """Returns spectrum of Vega loaded internally (always the same object)"""
    global __vega_spectrum
    if __vega_spectrum is None:
        __vega_spectrum = __get_new_vega_spectrum()
    return __vega_spectrum


def __get_new_vega_spectrum():
    from f311 import physics as ph
    return ft.load_spectrum(a99.get_path("data", "pysynphot-vega-fnu.xy", module=ph))


def get_vega_spectrum():
    """Returns a spectrum of the Vega star

    Creates a new `f311.filetypes.Spectrum` object at every call

    Returns:
        `f311.filetypes.Spectrum`: spectrum
    """
    return __get_new_vega_spectrum()


class Bandpass(object):
    """
    Wavelength filter band, with class tools

    This class is kept clean whereas UBVRIBands has examples and deeper documentation on parameters

    Args:
        tabular: ((wl, y), ...), 0 <= y <= 1
        parametric: ((wl, fwhm), ...)
        ref_mean_flux: reference mean flux passing through filter at magnitude 0 in Jy units
    """

    @property
    def l0(self):
        return self._get_l0()

    @property
    def lf(self):
        return self._get_lf()

    #     ****
    def __init__(self, name):
        self.name = name


    def _get_l0(self):
        raise NotImplementedError()


    def _get_lf(self):
        raise NotImplementedError


    def __mul__(self, other):
        raise TypeError("Bandpass left multiplication not defined")


    def __rmul__(self, other):
        """Right multiplication accepts Spectrum or tuple:(wave, flux)"""

        if isinstance(other, ft.Spectrum):
            out = copy.deepcopy(other)
            x, y = out.x, out.y
        elif isinstance(other, tuple) and len(other) == 2 and isinstance(other[0], np.ndarray) and \
             isinstance(other[1], np.ndarray) and len(other[0]) == len(other[1]):
            flag_spectrum = False
            out = copy.deepcopy(other)
            x, y = out
        else:
            raise TypeError("Invalid argument: "+str(other))

        # Overlap between bandpass filter and spectrum
        l0, lf = max(self.l0, x[0]), min(self.lf, x[-1])
        # Spectrum point indexes corresponding to this overlap (boolean mask)
        mask = np.logical_and(x >= l0, x <= lf)
        # Multiplies by filter
        y[mask] *= self.ufunc()(x[mask])
        # Points outside filter will be zero
        y[np.logical_not(mask)] = 0.
        return out


    def ufunc(self, flag_force_parametric=False):
        raise NotImplementedError()


    def area(self, l0, lf):
        """
        Calculates area (unit: a.u.*angstrom) under given range [l0, lf]
        Args:
            l0: lower edge of range
            lf: upper edge of range
        """
        # Calculates number of points as a fraction of REF_NUM_POINTS, minimum 10 points
        num_points = max(10, int((lf - l0) / (self.lf - self.l0) * _REF_NUM_POINTS))
        x = np.linspace(l0, lf, num_points)
        area = np.trapz(self.ufunc()(x), x)
        return area


class UBVTabulated(Bandpass):
    """
    Tabular filter

    Args:
        name: band name. Choices: U,B,V,R,I
    """

    # Tabular data for filters U,B,V,R,I
    # Michael Bessel 1990
    # Taken from http://spiff.rit.edu/classes/phys440/lectures/filters/filters.html
    DATA = collections.OrderedDict((
    ("U", ((3000, 0.00),  (3050, 0.016), (3100, 0.068), (3150, 0.167), (3200, 0.287), (3250, 0.423), (3300, 0.560),
           (3350, 0.673), (3400, 0.772), (3450, 0.841), (3500, 0.905), (3550, 0.943), (3600, 0.981), (3650, 0.993),
           (3700, 1.000), (3750, 0.989), (3800, 0.916), (3850, 0.804), (3900, 0.625), (3950, 0.423), (4000, 0.238),
           (4050, 0.114), (4100, 0.051), (4150, 0.019), (4200, 0.000))),
    ("B", ((3600, 0.0), (3700, 0.030), (3800, 0.134), (3900, 0.567), (4000, 0.920), (4100, 0.978), (4200, 1.000),
           (4300, 0.978), (4400, 0.935), (4500, 0.853), (4600, 0.740), (4700, 0.640), (4800, 0.536), (4900, 0.424),
           (5000, 0.325), (5100, 0.235), (5200, 0.150), (5300, 0.095), (5400, 0.043), (5500, 0.009), (5600, 0.0))),
    ("V", ((4700, 0.000), (4800, 0.030), (4900, 0.163), (5000, 0.458), (5100, 0.780), (5200, 0.967), (5300, 1.000),
          (5400, 0.973), (5500, 0.898), (5600, 0.792), (5700, 0.684), (5800, 0.574), (5900, 0.461), (6000, 0.359),
          (6100, 0.270), (6200, 0.197), (6300, 0.135), (6400, 0.081), (6500, 0.045), (6600, 0.025), (6700, 0.017),
          (6800, 0.013), (6900, 0.009), (7000, 0.000))),
    ("R", ((5500, 0.0), (5600, 0.23), (5700, 0.74), (5800, 0.91), (5900, 0.98), (6000, 1.000), (6100, 0.98),
           (6200, 0.96), (6300, 0.93), (6400, 0.90), (6500, 0.86), (6600, 0.81), (6700, 0.78), (6800, 0.72),
           (6900, 0.67), (7000, 0.61), (7100, 0.56), (7200, 0.51), (7300, 0.46), (7400, 0.40), (7500, 0.35),
           (8000, 0.14), (8500, 0.03), (9000, 0.00))),
    ("I", ((7000, 0.000), (7100, 0.024), (7200, 0.232), (7300, 0.555), (7400, 0.785), (7500, 0.910), (7600, 0.965),
           (7700, 0.985), (7800, 0.990), (7900, 0.995), (8000, 1.000), (8100, 1.000), (8200, 0.990), (8300, 0.980),
           (8400, 0.950), (8500, 0.910), (8600, 0.860), (8700, 0.750), (8800, 0.560), (8900, 0.330), (9000, 0.150),
           (9100, 0.030), (9200, 0.000)))
    ))

    def __init__(self, name):
        if name not in self.DATA:
            raise ValueError("Invalid band name, choices are ".format(*str(self.DATA.keys())))
        Bandpass.__init__(self, name)
        self.name = name
        self.x, self.y = zip(*self.DATA[name])

    def __repr__(self):
        return "UBVTabulated('{0!s}')".format(self.name)

    def _get_l0(self):
        return self.x[0]

    def _get_lf(self):
        return self.x[-1]

    def ufunc(self):
        return interp1d(self.x, self.y, kind='linear', bounds_error=False, fill_value=0)


class UBVParametric(Bandpass):
    """
    Parametric Gaussian filter

    Args:
        name: band name. Choices: U,B,V,R,I,Y,J,H,K,L,M,N,Q
        num_stds=3: number of standard deviations to left and right of midpoint to consider as
            band limits. Does not affect self.ufunc()
    """

    # Values taken from https://en.wikipedia.org/wiki/Photometric_system
    X0_FWHM = collections.OrderedDict((
    ("U", (3650., 660.)),
    ("B", (4450., 940.)),
    ("V", (5510., 880.)),
    ("R", (6580., 1380.)),
    ("I", (8060., 1490.)),
    ("Y", (10200., 1200.)),
    ("J", (12200., 2130.)),
    ("H", (16300., 3070.)),
    ("K", (21900., 3900.)),
    ("L", (34500., 4720.)),
    ("M", (47500., 4600.)),
    ("N", (105000., 25000.)),
    ("Q", (210000., 58000.))
    ))

    def __init__(self, name, num_stds=3):
        if name not in self.X0_FWHM:
            raise ValueError("Invalid band name, choices are ".format(*str(self.X0_FWHM.keys())))
        Bandpass.__init__(self, name)
        self.x0, self.fwhm = self.X0_FWHM[name]
        self.num_stds = num_stds

        std = self.fwhm * (1. / np.sqrt(8 * np.log(2)))
        self._l0, self._lf = self.x0 - num_stds * std, self.x0 + num_stds * std

    def __repr__(self):
        return "UBVParametric('{0!s}', {1:d})".format(self.name, self.num_stds)

    def _get_l0(self):
        return self._l0

    def _get_lf(self):
        return self._lf

    def ufunc(self):
        return _ufunc_gauss(self.x0, self.fwhm)


def _ufunc_gauss(x0, fwhm):
    """
    Returns Gaussian function (callable) that works as a numpy ufunc

    Args:
        x0: maximum value, where the function evaluates to 1.
        fwhm: "Full-Width at Half-Maximum"

    Reference: https://en.wikipedia.org/wiki/Gaussian_function
    """
    K = 4 * np.log(2) / fwhm ** 2
    def f(x):
        return np.exp(-(x - x0) ** 2 * K)
    return f


# Standard flux, according to several references
#
# TODO get references: Bessel etc
#
# values taken from https://en.wikipedia.org/wiki/Apparent_magnitude
# I- and J-band values agree with Evans, C.J., et al., A&A 527(2011): A50.
#
# Unit is erg/cm**2/s/Hz
STDFLUX = collections.OrderedDict((
("U", 1.81e-20),
("B", 4.26e-20),
("V", 3.64e-20),
("R", 3.08e-20),
("I", 2.55e-20),
("J", 1.60e-20),
("H", 1.08e-20),
("K", 0.67e-20),
))


def __get_ubv_bandpass(bp):
    """Internal function to convert string to Bandpass, or return the bp if already a Bandpass"""
    if isinstance(bp, Bandpass):
        pass
    elif isinstance(bp, str):
        if bp in "UBVRI":
            bp = UBVTabulated(bp)
        else:
            bp = UBVParametric(bp)
    else:
        raise ValueError("Invalid value for argument 'bandpass': {0!s}".format(str(bp)))
    return bp



"""
Multiplicity (Chemistry)

https://en.wikipedia.org/wiki/Multiplicity_(chemistry)

  - Formulas to calculate the Hönl-London factors
  - Conversion from quantum information to branch label such as P/Q/R/P1/Q1 etc.

Note: This package will preserve the namespace hierarchy because there are repeated method names, e.g.,
`quanta_to_branch()`
"""



_J2L_MAP = {"P": -1, "Q": 0, "R": 1}
def branch_to_J2l(br):
    """Converts P/Q/R to -1/0/+1

    Tolerant to strings such as "P1", "P12" etc. (i.e., takes only the first character)"""
    return _J2L_MAP[br[0]]


from . import singlet
from . import doublet



# # Temporary imports
#   =================
# These modules should be be del'eted at the end
import a99


# # Setup
#   =====
SESSION_PREFIX_SINGULAR = 'session-'
SESSION_PREFIX_PLURAL = 'session-'
MULTISESSION_PREFIX = 'multi-session-'


# # Imports
#   =======
from .errors import *
from .conf import *
from .runnables import *
from .rm import *
from .util import *
from .multirunnable import *
from .gui import *
from .paths import *
from . import gui


# # Function to access package-specific config file
#   ===============================================
def get_config():
    """Returns PyfantConfigObj object that corresponds to file ~/.ftpyfant.conf"""
    return a99.get_config_obj(".f311.pyfant.conf")


# # Finally, gets rid of unwanted symbols in the workspace
#   ======================================================
del a99



__all__ = ["ToScalar_SNR", "ToScalar_Magnitude", "ToScalar_UseNumPyFunc"]


from .basic import ToScalar
import numpy as np
import f311.physics as ph

class ToScalar_SNR(ToScalar):
    """
    Calculates Signal-to-noise ratio (SNR) using a part of the "signal" (i.e. the spectrum)

    Formula: SNR = sqrt((y_**2)) / std(y_),
             where:
                 y_ is the  slice of of the spectrum flux vector within the range [llzero, llfin];
                 the numerator is the RMS value of y_

    **Note** this has been tested to be consistent with IRAF SNR calculation

    **Note** It is assumed that the "signal" is *stationary* within [llzero, llfin]
    meaning that the mean and variance of the "signal" is the same for all points within
    this region (more precisely "weak-sense stationary"
    (https://en.wikipedia.org/wiki/Stationary_process#Weak_or_wide-sense_stationarity))
    """

    def __init__(self, llzero, llfin):
        ToScalar.__init__(self)
        self.llzero = llzero
        self.llfin = llfin

    def _do_use(self, inp):
        x = inp.x
        y = inp.y
        signal = y[np.logical_and(x >= self.llzero, x <= self.llfin)]

        a = np.sqrt(np.mean(signal**2))
        b = np.std(signal)

        if a == 0 and b == 0:
            output = float("nan")
        elif b == 0:
            output = float("inf")
        else:
            output = a/b
        return output


class ToScalar_Magnitude(ToScalar):
    """
    Calculates the magnitude of a spectrum

        Args:
            band_name: U/B/V/R/I/Y/J/H/K/L/M/N/Q
            system: reference magnitude system.
                Choices:
                    "stdflux": literature reference values for bands U,B,V,R,I,J,H,K only
                    "vega": uses the Vega star spectrum as a reference
                    "ab": AB[solute] magnitude system
            zero_point: subtracts this value from the calculated magnitude to implement some desired
                        correction.
            flag_force_band_range: (default: False) if set, will consider that the spectrum
                                   extends over the full range of the band even if it is narrower
                                   than that
    """


    def __init__(self, band_name, system="stdflux", zero_point=0., flag_force_band_range=False):
        ToScalar.__init__(self)
        self.band_name = band_name
        self.system = system
        self.zero_point = zero_point
        self.flag_force_band_range = flag_force_band_range

    def _do_use(self, inp):
        return ph.calc_mag(inp, self.band_name, self.system, self.zero_point,
                           self.flag_force_band_range)


class ToScalar_UseNumPyFunc(ToScalar):
    """
    Reduces spectrum y-vector to scalar using a numpy function, e.g., np.mean(), np.std()
    """

    def __init__(self, func=np.mean):
        ToScalar.__init__(self)
        self.func = func

    def _do_use(self, inp):
        out = self.func(inp.y)
        if isinstance(out, np.ndarray):
            try:
                # tries nice string representation;
                s = self.func.__name__
            except:
                # if fails, goes for generic string conversion
                s = str(self.func)
            raise RuntimeError("Function '{}' does not evaluate to scalar".format(s))
        return out


__all__ = ["XHelpDialog"]


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
from .... import explorer as ex
import f311.filetypes as ft


class XHelpDialog(a99.XLogDialog):
    """
    Dialog with two areas: Fields Area and Help Area

    1. The Fields area contains a grid layout and the OK & Cancel buttons
    2. The Help Area contains a combo box and a text area

    Relevant attributes:
      self.help_data: [(class name without prefix, docstring), ...]
      self.block: None or ToScalar instance, set before closing when one clicks on "OK"
      self.grid: grid layout (initially empty)
      self.labelHelpTopics: label with text "Help Topics", exposed in case you want to change this text
      self.comboBox: combo box to add the help topics

    """

    def __init__(self, *args):
        a99.XLogDialog.__init__(self, *args)

        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        self.help_data = []

        # # Central layout
        lymain = self.centralLayout = QVBoxLayout()
        a99.set_margin(lymain, 0)
        self.setLayout(lymain)

        # ## Title of Fields Area
        label = keep_ref(QLabel("<b>Setup</b>"))
        lymain.addWidget(label)

        # ## Fields Area
        wfields = QWidget()
        wfields.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        lyfields = keep_ref(QHBoxLayout(wfields))
        lymain.addWidget(wfields)
        a99.set_margin(lyfields, 0)
        lyfields.setSpacing(2)

        # ### Left area of Fields Area
        lypanel = self.grid = QGridLayout()
        lyfields.addLayout(lypanel)


        # ### Right area of Fields Area (button box)
        bb = keep_ref(QDialogButtonBox())
        bb.setOrientation(Qt.Vertical)
        bb.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        lyfields.addWidget(bb)
        bb.rejected.connect(self.reject)
        bb.accepted.connect(self.accept)

        # ## Title of Help Area
        label = keep_ref(QLabel("<b>Help</b>"))
        lymain.addWidget(label)

        # ## Help Area

        # ### Panel containing a combobox to choose among help topics
        lypanel = QHBoxLayout()
        lymain.addLayout(lypanel)
        a99.set_margin(lypanel, 0)
        lypanel.setSpacing(2)
        ###
        label = self.labelHelpTopics = QLabel("Help Topics")
        lypanel.addWidget(self.labelHelpTopics)
        ###
        cb = self.comboBox = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_changed)
        lypanel.addWidget(cb)
        label.setBuddy(cb)
        ###
        lypanel.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # ### Text Edit below combobox
        x = self.textEdit = QTextEdit()
        x.setReadOnly(True)
        x.setStyleSheet("QTextEdit {color: %s}" % a99.COLOR_DESCR)
        lymain.addWidget(x)

    def combobox_changed(self):
        idx = self.comboBox.currentIndex()
        # self.textEdit.setText("<h1>%s</h1>%s\n" % (self.help_data[idx][0], self.help_data[idx][1].replace("\n", "<br>")))


        self.textEdit.setText("<h3>{}{}</h3><pre>{}</pre>\n".format(*self.help_data[idx]))





"""
Miscellanea routines that depend on other pyfant modules.

Rule: only 'gui/' modules can import util!!!

"""
import os
import glob
import shutil
import a99
from .. import pyfant as pf
from .. import filetypes as ft


__all__ = [
    "run_parallel", "setup_inputs", "copy_star", "link_to_data", "create_or_replace_or_skip_links",
    "copy_or_skip_files",
]



# ##################################################################################################
# Terminal-based interface


# TODO TEST run_parallel
def run_parallel(rr, max_simultaneous=None, flag_console=True, runnable_manager=None):
    """
    Args:
      rr: list of Runnable instances
      max_simultaneous: (optional, default is RunnableManager default)
       maximum number of simultaneous processes.
      runnable_manager=None: (optional) if passed, will use passed;
       if not, will create new.

    Returns: the RunnableManager object
    """

    # Adds to pool
    logger = a99.get_python_logger()
    if runnable_manager:
        assert isinstance(runnable_manager, pf.RunnableManager)
        rm = runnable_manager
    else:
        rm = pf.RunnableManager(max_simultaneous=max_simultaneous)
    flag_had_to_start = False
    if not rm.flag_start_called:
        rm.start()
        flag_had_to_start = True

    rm.add_runnables(rr)

    # Primitive thread monitor
    if flag_console:
        while True:
            print(("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" + (
            " ALIVE" if rm.is_alive() else " DEAD")))
            print(rm)
            print(("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" + (
            " ALIVE" if rm.is_alive() else " DEAD")))
            s = input("[Enter] -- [e]xit & keep in loop -- [q]uit -- [k]ill running >>> ")
            if s.lower() == "q":
                if rm.is_alive():
                    try:
                        rm.exit()
                    except:
                        logger.exception("Error trying to exit")
                break
            if s.lower() == "e":
                try:
                    rm.exit()
                except:
                    logger.exception("Error trying to exit")
            if s.lower() == "k":
                rm.kill_runnables()
    else:
        rm.wait_until_finished()
        if flag_had_to_start:
            rm.exit()

    a99.get_python_logger().info(
        ("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" + (" ALIVE" if rm.is_alive() else " DEAD")))
    a99.get_python_logger().info("test-tm2 [SUPPOSED TO HAVE] EXITED")
    a99.get_python_logger().info(rm)

    return rm


def setup_inputs(dest_dir='.', star='sun-asplund-2009', common='common', h=True, atoms=True,
                 molecules=True, opa=True):
    """
    Sets up input data for spectral synthesis.

    Args:
      dest_dir='.': directory where files and links will be created
      star='sun-asplund-2009': directory (relative to PFANT/data) for stellar data
      common='common': directory (relative to PFANT/data) for star-independent data files
      h=True: whether to look for hmap.dat
      atoms=True: whether to look for atoms.dat
      molecules=True: whether to look for molecules.dat
      opa=True: whether to look for grid.moo
    """

    logger = a99.get_python_logger()
    dd = pf.get_pfant_path("data")

    # Functions that return full path, given a filename, to ...
    fd = lambda filename: os.path.join(dest_dir, filename)  # ... Destination directory
    fs = lambda filename: os.path.join(dd, star, filename)  # ... Stellar data directory
    fc = lambda filename: os.path.join(dd, common, filename)  # ... Common data directory

    # ## main.dat not present
    if not os.path.isfile(fd("main.dat")):
        zz_mnbp = ["main.dat", "abonds.dat",
                   "modeles.mod"]  # files that must not be present if main.dat is not present

        for z in zz_mnbp:
            if os.path.isfile(fd(z)):
                raise RuntimeError("Found file '%s' in local directory."
                                   "If 'main.dat' is not present, files %s must also not exist." % zz_mnbp[
                                                                                                   1:])

    # ## Stellar data...
    zz = ["main.dat", "abonds.dat"]
    copy_or_skip_files([fs(z) for z in zz], dest_dir=dest_dir)

    # ## Common data...
    zz = ["absoru2.dat", "partit.dat", "grid.mod"]
    if opa: zz.append("grid.moo")
    if h: zz.append("hmap.dat")
    if atoms: zz.append("atoms.dat")
    if molecules: zz.append("molecules.dat")
    create_or_replace_or_skip_links([fc(z) for z in zz], dest_dir=dest_dir)


def copy_star(src_dir):
    star_classes = [ft.FileMain, ft.FileDissoc, ft.FileAbonds]

    print(("Will look inside directory %s" % src_dir))

    # makes list of files to analyse
    types = ('*.dat', '*.mod')
    ff = []
    for type_ in types:
        ff.extend(glob.glob(os.path.join(src_dir, type_)))

    copy_or_skip_files(ff)


def link_to_data(src_dir):
    star_classes = [ft.FileMain, ft.FileDissoc, ft.FileAbonds]

    print(("Will look inside directory %s" % src_dir))

    # makes list of files to analyse
    types = ('*.dat', '*.mod', '*.moo')
    ff = []
    for type_ in types:
        ff.extend(glob.glob(os.path.join(src_dir, type_)))

    create_or_replace_or_skip_links(ff)


def create_or_replace_or_skip_links(ff, dest_dir="."):
    """Creates a series of links given a list of target filepaths.

    Args:
      ff: list of full path to files
      dest_dir=".": destination directory

    It skips files of types FileMain, FileAbonds, FileDissoc, FileToH
    """
    for f in ff:
        name = os.path.split(f)[1]
        ptd = os.path.join(dest_dir, name)  # path to destination

        flag_skip = False
        print(("Considering file '%s' ..." % name))
        if os.path.isfile(ptd) and not os.path.islink(ptd):
            a99.print_skipped("file exists in local directory")
            flag_skip = True
        else:
            obj = a99.load_with_classes(f, [ft.FileMain, ft.FileAbonds, ft.FileDissoc, ft.FileToH])
            if obj is not None:
                a99.print_skipped("detected type %s" % obj.__class__.__name__)
                flag_skip = True
            else:
                obj = a99.load_with_classes(f, [ft.FileModBin])
                if obj is not None:
                    if len(obj) == 1:
                        a99.print_skipped("%s of only one record" % obj.__class__.__name__)
                        flag_skip = True

        if not flag_skip:
            try:
                if os.path.islink(ptd):
                    os.remove(ptd)
                    s_action = "replaced existing"
                else:
                    s_action = "created"
                a99.create_symlink(f, ptd)
                print(("   ... %s link" % s_action))
            except Exception as e:
                a99.print_error("Error creating link: %s" % str(e))


def copy_or_skip_files(ff, dest_dir="."):
    """Copies a series of files, skipping those which already exist.

    Args:
      ff: list of full paths to files to be copied
      dest_dir=".": destination directory
    """

    for f in ff:
        name = os.path.split(f)[1]

        flag_skip = False
        print(("Considering file '%s' ..." % name))
        if os.path.isfile(name):
            a99.print_skipped("file exists in local directory")
            flag_skip = True
        else:
            obj = a99.load_with_classes(f, [ft.FileMain, ft.FileAbonds, ft.FileDissoc])
            if obj is not None:
                pass
            else:
                a99.print_skipped("neither main, abonds, nor dissoc file")
                flag_skip = True

        if not flag_skip:
            try:
                shutil.copy(f, dest_dir)
                print("   ... file copied")
            except Exception as e:
                a99.print_error("Error copying file: %s" % str(e))




import a99
from .. import get_hitran_molecules, get_hitran_isotopologues
import f311.filetypes as ft

__all__ = ["populate_hitrandb"]


def populate_hitrandb(db):
    """Populates database with HITRAN data

    Populates a sqlite3 database represented by a FileHitranDB object with information downloaded
    from the HITRAN website

    Args:
        db: FileHitranDB instance
    """

    assert isinstance(db, ft.FileHitranDB)
    conn = db.get_conn()
    try:

        mols, _ = get_hitran_molecules()

        conn.executemany("insert into molecule values (?,?,?)", mols)

        a99.get_python_logger().info("Inserted {} molecules".format(len(mols)))

        for mol in mols:
            isos, _ = get_hitran_isotopologues(mol[0])

            for iso in isos:
                try:
                    conn.execute("insert into isotopologue values(?,?,?,?,?)",
                                 [iso[0], mol[0]] + iso[1:])
                except:
                    a99.get_python_logger().exception("Tried to insert this: {}".format(iso))
                    raise

            a99.get_python_logger().info("Inserted {} isotopologues for molecule '{}' ({})".
                                         format(len(isos), *mol[1:3]))
    finally:
        conn.commit()
        conn.close()

    conn.close()



import a99
import f311.filetypes as ft
from ... import convmol as cm
from ... import pyfant as pf

__all__ = ["populate_moldb"]


def populate_moldb(moldb):
    """Tries to download data from NIST Web Book online"""

    assert isinstance(moldb, ft.FileMolDB)

    conn = moldb.get_conn()
    # TODO gather more formulae
    formulae = ["MgH", "C2", "CN", "CH", "NH", "CO", "OH", "FeH", "TiO"]

    # Uses PFANT/data/common/molecules.dat to retrieve "fe", "do", "am", etc.
    filemol = ft.FileMolecules()
    filemol.load(pf.get_pfant_data_path("common", "molecules.dat"))
    bysym = dict([(tuple(m.symbols), m) for m in filemol])

    for formula in formulae:
        try:
            data, _, name = cm.get_nist_webbook_constants(formula)

            fe, do, am, bm, ua, ub, te, cro, s = None, None, None, None, None, None, None, None, \
                                                 None

            # Tries to retrieve "fe", "do" etc from molecules.dat
            symbols = ft.description_to_symbols(formula)
            if not symbols:
                a99.get_python_logger().warning("Formula '{}' not in internal BUILTIN_FORMULAS, "
                                                "and probably this molecule is also not in PFANT/data/common/molecules.dat".
                                                format(formula))

                symbols = ["", ""]
            else:
                m = bysym.get(tuple(symbols))
                if m:
                    fe = m.fe
                    do = m.do
                    am = m.am
                    bm = m.bm
                    ua = m.ua
                    ub = m.ub
                    te = m.te
                    cro = m.cro
                    s = m.s

            symbols = [x.strip() for x in symbols]
            conn.execute("insert into molecule "
                         "(formula, name, symbol_a, symbol_b, fe, do, am, bm, ua, ub, te, cro, s) "
                         "values (?,?,?,?,?,?,?,?,?,?,?,?,?)",
                         (
                         formula, name, symbols[0], symbols[1], fe, do, am, bm, ua, ub, te, cro, s))

            id_molecule = conn.execute("select last_insert_rowid() as id").fetchone()["id"]
            for state in data:
                # **Note** assumes that the columns in data match the
                # (number of columns in the state table - 2) and their order
                conn.execute("insert into state values (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)",
                             [None, id_molecule] + state)

            conn.commit()

        except:
            a99.get_python_logger().exception("Failed for molecule '{}'".format(formula))

    conn.close()



#!/usr/bin/env python3


"""Downloads and prints molecular constants from NIST Web Book for a particular molecule

Example:

    print-nist.py OH

"""


import tabulate
import sys
import a99
import logging
import argparse
import f311.convmol as cm


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=a99.SmartFormatter)
    parser.add_argument('formula', type=str, help='NIST formula', nargs=1)
    args = parser.parse_args()

    data, header, title = cm.get_nist_webbook_constants(args.formula)
    print("\n*** {} ***\n".format(title))
    print(tabulate.tabulate(data, header))





from .datafile import DataFile
import a99
import os
import re

__all__ = ["FilePy"]


class FilePy(DataFile):
    """
    Configuration file saved as a .py Python source script

    This class is not intendend to be instantialized. It is an ancestor class for other classes.
    """

    def _test_magic(self, filename):
        with open(filename, "r") as file:
            line = file.readline()
            if not re.match("\s*#\s*-\*-\s*FilePy:\s*{}\s*-\*-".format(self.classname), line):
                raise RuntimeError("File '{}' does not appear to be a '{}' (expected first line of code:"
                                   " \"{}\")".format(filename, self.classname, self._get_magic()))

    def _get_magic(self):
        """Returns string to be written as first line of .py file

         **Note** Newline character **not** included"""

        return "# -*- FilePy: {} -*-".format(self.classname)


__all__ = ["air_to_vacuum", "vacuum_to_air"]


def vacuum_to_air(λvac):
    """
    Vacuum-to-air wavelength conversion

    Args:
        λvac: vacuum wavelength in Angstrom

    Returns:
        λair: air wavelength in Angstrom

    Source:
        http://www.astro.uu.se/valdwiki/Air-to-vacuum%20conversion

    Original reference:
        Donald Morton (2000, ApJ. Suppl., 130, 403)

    Example:

    >>> # The following code converts wavenumber=2000 cm**-1 to wavelength in Angstrom
    >>> vacuum_to_air(1e8/2000)
    49986.36934549974
    """
    # if λvac <= 0:
    #     raise ValueError("λvac must be > 0")

    s = 1e4 / λvac
    n = 1 + 0.0000834254 + 0.02406147 / (130 - s ** 2) + 0.00015998 / (38.9 - s ** 2)
    λair = λvac / n
    return λair


def air_to_vacuum(λair):
    """
    Air-to-vacuum wavelength conversion

    Args:
        λair: air wavelength in Angstrom

    Returns:
        λvac: vacuum wavelength in Angstrom

    Formula author (according to [1]):
        N. Piskunov

    Example:

    >>> # The following code converts wavelength=5500 Angstrom to wavenumber in cm**-1:
    >>> 1e8/air_to_vacuum(5500.)
    18176.768046090445

    Example:

    >>> # The following code reproduces the figure shown in [1] ("comparison of the Morton and the inverse
    >>> # transformation by NP between 2000 Å and 100000 Å.")
    >>> import matplotlib.pyplot as plt
    >>> import numpy as np
    >>> λvac = 10**np.linspace(np.log10(2000), np.log10(1000000), 2000)
    >>> y = air_to_vacuum(vacuum_to_air(λvac))-λvac
    >>> _ = plt.semilogx(λvac, y)
    >>> _ = plt.xlabel("$\lambda$ in Angstroem")
    >>> _ = plt.ylabel("$\Delta\lambda$")
    >>> _ = plt.xlim([λvac[0]-50, λvac[-1]])
    >>> _ = plt.title("air_to_vacuum(vacuum_to_air($\lambda_{vac}$))-$\lambda_{vac}$")
    >>> _ = plt.tight_layout()
    >>> _ = plt.show()

    Source:
        [1] http://www.astro.uu.se/valdwiki/Air-to-vacuum%20conversion
    """
    # if λair <= 0:
    #     raise ValueError("λair must be > 0")

    s = 1e4 / λair
    n = 1 + 0.00008336624212083 + 0.02408926869968 / (130.1065924522 - s**2) + 0.0001599740894897 / \
        (38.92568793293 - s**2)
    λvac = λair * n
    return λvac



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
# import f311.ftpyfant as pf
# from a_WState import WState
# import moldb as db
from .a_WDBState import WDBState


__all__ = ["WStateConst"]


class WStateConst(a99.WBase):
    """
    Widget for the user to select/modify state-specific constants. Has a embedded WDBState
    """

    @property
    def fieldnames(self):
        return self._fieldnames

    @property
    def constants(self):
        """Returns a dictionary with all molecular constants"""
        ret = {}
        for fieldname in self._fieldnames:
            ret[fieldname] = self[fieldname]
        return ret

    @property
    def row(self):
        """Wraps WDBMolecule.row"""
        return self.w_mol.row

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self.flag_populating = False  # activated when populating table

        # # Internal state

        # Molecular constants of interest
        self._fieldnames = ["omega_e", "omega_ex_e", "omega_ey_e", "B_e", "alpha_e", "D_e",
                            "beta_e", ]

        # dictionary {(field name): (edit object), }
        # (will be populated later below together with edit widgets creation)
        self._edit_map = {}

        # # GUI design

        l = QVBoxLayout(self)
        a99.set_margin(l, 1)

        # ## State widget
        a = self.w_state = WDBState(self.parent_form)
        l.addWidget(a)
        a.id_changed.connect(self._w_state_id_changed)

        # ## Title label
        a = self.apa = self.keep_ref(QLabel("<h4>Values to be used in calculations</h4>"))
        l.addWidget(a)
        a.setToolTip(
            "If you values below, this will affect the conversion calculation, but the information will not be written back to the database")

        # ## Molecular constants edit fields

        lg = QGridLayout()
        l.addLayout(lg)
        nr, nc, n = 3, 3, len(self._fieldnames)
        ti = a99.get_table_info("moldb", "state")
        for j in range(nc):
            # ### One grid layout for each column of fields
            ii = range(j, n, nc)
            for i in range(len(ii)):
                fieldname = self._fieldnames[ii[i]]
                info = ti[fieldname]
                caption = info["caption"] or fieldname
                a = QLabel(caption)
                e = QLineEdit("")
                tooltip = info["tooltip"]
                if tooltip:
                    a.setToolTip(tooltip)
                    e.setToolTip(tooltip)
                self._edit_map[fieldname] = e
                lg.addWidget(a, i, j * 2)
                lg.addWidget(e, i, j * 2 + 1)

    def __getitem__(self, fieldname):
        """Allows dict-like access of molecular constants of interest. Returns float value or None"""
        if fieldname in self._fieldnames:
            try:
                return float(self._edit_map[fieldname].text())
            except ValueError:
                return None

    def None_to_zero(self):
        """Fills missing values with zeros"""
        for edit in self._edit_map.values():
            if len(edit.text().strip()) == 0:
                edit.setText("0")

    def set_id_molecule(self, id_):
        """Calls WDBState.set_id_molecule"""
        self.w_state.set_id_molecule(id_)

    def validate(self):
        """Returns True if all edit fields can be converted to float, otherwise False"""
        ret = True
        for fieldname in self._fieldnames:
            try:
                _ = float(self[fieldname])
            except ValueError:
                self.add_log_error("Failed converting field '{}' to float".format(fieldname))
                ret = False
                break
        return ret

    def _w_state_id_changed(self):
        """Transfer values from self.data to edit fields below the table widget"""
        # id_ = print("State id changed to ", self.w_state.id)
        row = self.w_state.row
        # if not row:
        #     return
        for fieldname, edit in self._edit_map.items():
            value = "" if not row else row[fieldname]
            edit.setText(str(value) if value is not None else "")



"""
Miscellanea routines that depend on other modules and sub-packages.
"""

import logging
import a99
# from f311 import filetypes as ft
from astropy.io import fits

# TODO usage example for list_data_types(), or even a script

__all__ = [
    "load_any_file", "load_spectrum", "load_spectrum_fits_messed_x", "list_data_types",
]

def load_any_file(filename):
    """
    Attempts to load filename by trial-and-error using _classes as list of classes.

    Returns:
        A DataFile descendant, whose specific class depends on the file format detected, or None
        if the file canonot be loaded
    """
    from f311 import filetypes as ft

    # Splits attempts using ((binary X text) file) criterion
    if a99.is_text_file(filename):
        return a99.load_with_classes(filename, ft.classes_txt())
    else:
        return a99.load_with_classes(filename, ft.classes_bin())


def load_spectrum(filename):
    """
    Attempts to load spectrum as one of the supported types.

    Returns:
        a Spectrum, or None
    """
    from f311 import filetypes as ft

    f = a99.load_with_classes(filename, ft.classes_sp())
    if f:
        return f.spectrum
    return None


def load_spectrum_fits_messed_x(filename, sp_ref=None):
    """Loads FITS file spectrum that does not have the proper headers. Returns a Spectrum"""
    from f311 import filetypes as ft

    # First tries to load as usual
    f = a99.load_with_classes(filename, (ft.FileSpectrumFits,))

    if f is not None:
        ret = f.spectrum
    else:
        hdul = fits.open(filename)

        hdu = hdul[0]
        if not hdu.header.get("CDELT1"):
            hdu.header["CDELT1"] = 1 if sp_ref is None else sp_ref.delta_lambda
        if not hdu.header.get("CRVAL1"):
            hdu.header["CRVAL1"] = 0 if sp_ref is None else sp_ref.x[0]

        ret = ft.Spectrum()
        ret.from_hdu(hdu)
        ret.filename = filename
        original_shape = ret.y.shape  # Shape of data before squeeze
        # Squeezes to make data of shape e.g. (1, 1, 122) into (122,)
        ret.y = ret.y.squeeze()

        if len(ret.y.shape) > 1:
            raise RuntimeError(
                "Data contains more than 1 dimension (shape is {0!s}), "
                "FITS file is not single spectrum".format(original_shape))

    return ret


def list_data_types():
    """
    Returns a list with all data types, in Markdown table format
    """
    from f311 import filetypes as ft
    ll = []  # [(description, default filename), ...]

    for attr in ft.classes_file():
        doc = attr.__doc__
        doc = attr.__name__ if doc is None else doc.strip().split("\n")[0]

        def_ = attr.default_filename
        def_ = def_ if def_ is not None else "-"
        ll.append((doc, def_))

    ll.sort(key=lambda x: x[0])

    return a99.markdown_table(("File type", "Default filename (for all purposes)"), ll)



#!/usr/bin/env python3

"""
Creates HTML reports from WebSim-COMPASS output files
"""


import sys
import argparse
import logging
import glob
import os
import logging
from f311 import aosss
import a99


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":
    lggr = a99.get_python_logger()

    parser = argparse.ArgumentParser(
     description=__doc__,
     formatter_class=a99.SmartFormatter
    )

    parser.add_argument('--dir', nargs='?', default='.', type=str,
     help='Input directory')
    parser.add_argument('--max', metavar='N', type=int, default=100,
                        help='Maximum allowed number of reports')
    parser.add_argument('numbers', metavar='N', type=str, nargs='+',
                     help='List of simulation numbers (single value and ranges accepted, e.g. 1004, 1004-1040)')

    args = parser.parse_args()

    simids0 = aosss.compile_simids(args.numbers)

    # Checks which ones actually exist in directory
    simids = []
    for simid in simids0:
        ff = glob.glob(os.path.join(args.dir, "%s*" % simid))
        if len(ff) > 0:
            simids.append(simid)

    lggr.info("List of simulation IDs: " + ", ".join(simids))

    if len(simids) > args.max:
        lggr.critical("Number of simulations to report (%d) exceeds "
                                            "maximum number (%d)" % (len(simids), args.max))
        lggr.critical("Use --max option to raise this maximum number")
        sys.exit()

    for simid in simids:
        lggr.info("Creating report for simulation '{}'...".format(simid))
        fn_output = aosss.create_simulation_report(simid, args.dir)
        lggr.info("File '%s' created successfully" % fn_output)



""""
List of spectra sharing same wavenumber axis. Uses FITS format
"""
# from a99 import Spectrum, AttrsPart, froze_it, eval_fieldnames
import a99
from astropy.io import fits
import os
import numpy as np
import numbers
import copy
# import f311.physics as ph
# import f311.filetypes as ft
from .. import Spectrum

__all__ = ["SpectrumCollection"]


class SpectrumCollection(a99.AttrsPart):
    """Base class, maintains spectra with "more headers"; to/from HDU without much interpretation"""

    attrs = ["fieldnames", "more_headers", "spectra"]

    def __init__(self):
        a99.AttrsPart.__init__(self)
        self._flag_created_by_block = False  # assertion
        self.filename = None
        # _LambdaReference instance, can be inferred from first spectrum added
        self.spectra = []
        # List of field names
        self.fieldnames = []
        # Visible field names (GUI setup purpose); order is same as column order in table widget
        self.fieldnames_visible = []
        self.more_headers = {}

    def __len__(self):
        return len(self.spectra)

    # def __getitem__(self, item):
    #     """Return copy of self with sliced self.spectra"""
    #     ret = copy.copy(self)
    #     ret.spectra = self.spectra.__getitem__(item)

    # - NAXIS(1/2/3) apparently managed by pyfits
    # - FIELDNAM & FIELDN_V are parsed separately
    _IGNORE_HEADERS = ("NAXIS", "FIELDNAM", "FIELDN_V")

    def from_hdulist(self, hdul):
        assert isinstance(hdul, fits.HDUList)
        if not (hdul[0].header.get("ANCHOVA") or hdul[0].header.get("TAINHA")):
            raise RuntimeError("Wrong HDUList")

        self.spectra = []
        for i, hdu in enumerate(hdul):
            if i == 0:
                # Additional header fields
                for name in hdu.header:
                    if not name.startswith(self._IGNORE_HEADERS):
                        self.more_headers[name] = hdu.header[name]

                # self.fieldnames is not overwritten if there is no such information in HDU
                temp = hdu.header.get("FIELDNAM")
                if temp:
                    self.fieldnames = a99.eval_fieldnames(temp)
                # self.fieldnames_visible is overwritten yes
                temp = hdu.header.get("FIELDN_V")
                if temp is None:
                    self.fieldnames_visible = copy.copy(self.fieldnames)
                else:
                    self.fieldnames_visible = a99.eval_fieldnames(temp)
            else:
                sp = Spectrum()
                sp.from_hdu(hdu)
                self.add_spectrum(sp)

    def to_hdulist(self):
        # I think this is not or should not be a restriction assert len(self.spectra) > 0, "No spectra added"

        # dl = self.delta_lambda

        hdul = fits.HDUList()

        hdu = fits.PrimaryHDU()
        hdu.header["FIELDNAM"] = str(self.fieldnames)
        hdu.header["FIELDN_V"] = str(self.fieldnames_visible)
        hdu.header["ANCHOVA"] = 26.9752

        hdu.header.update(self.more_headers)
        hdul.append(hdu)

        for sp in self.spectra:
            hdul.append(sp.to_hdu())

        return hdul

    def collect_fieldnames(self):
        """Returns a list of unique field names union'ing all spectra field names"""
        # self.fieldnames = []
        ff = []
        for sp in self.spectra:
            ff.extend(list(sp.more_headers.keys()))
        return list(set(ff))

    def add_spectrum(self, sp):
        """Adds spectrum, no content check

        Updates self.fieldnames to expose sp.more_headers

        Nullifies spectrum filename and assigns the file basename to more_headers["ORIGIN"]
        """
        assert isinstance(sp, Spectrum)

        if sp.filename is not None:
            sp.more_headers["ORIGIN"] = os.path.basename(sp.filename)
            sp.filename = None
        self.spectra.append(sp)

        for name in sp.more_headers:
            if name not in self.fieldnames:
                self.fieldnames.append(name)

    def delete_spectra(self, indexes):
        """Deletes spectra given a list of 0-based indexes"""
        if isinstance(indexes, numbers.Integral):
            indexes = [indexes]
        indexes = list(set(indexes))
        n = len(self.spectra)
        if any([idx < 0 or idx >= n for idx in indexes]):
            raise RuntimeError("All indexes must be (0 le index lt %d)" % n)
        for index in reversed(indexes):
            del self.spectra[index]

    def clear(self):
        """Removes all spectra from the collection"""
        self.spectra = []

    def merge_with(self, other):
        """Adds spectra from other SpectrumCollection to self"""
        assert isinstance(other, SpectrumCollection)
        for sp in other.spectra:
            self.add_spectrum(sp)


    # def to_colors(self, visible_range=None, flag_scale=False, method=0):
    #     """Returns a [n, 3] red-green-blue (0.-1.) matrix
    #
    #     Args:
    #       visible_range=None: if passed, the true human visible range will be
    #                             affine-transformed to visible_range in order
    #                             to use the red-to-blue scale to paint the pixels
    #       flag_scale: whether to scale the luminosities proportionally
    #                     the weight for each spectra will be the area under the flux
    #       method: see Spectrum.get_rgb()
    #     """
    #     weights = np.zeros((len(self), 3))
    #     max_area = 0.
    #     ret = np.zeros((len(self), 3))
    #     for i, sp in enumerate(self.spectra):
    #         ret[i, :] = ph.spectrum_to_rgb(sp, visible_range, method)
    #         sp_area = np.sum(sp.y)
    #         max_area = max(max_area, sp_area)
    #         weights[i, :] = sp_area
    #     if flag_scale:
    #         weights *= 1. / max_area
    #         ret *= weights
    #     # TODO return weights if necessary
    #     return ret



__all__ = ["XGroupSpectra"]


from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_XHelpDialog import *
import a99
from .... import explorer as ex
import f311.filetypes as ft


class XGroupSpectra(XHelpDialog):
    """
    Edit Parameters gb.py a SpectrumList

    Relevant attributes (set on close):
      self.block: None or GroupBlock instance
      self.group_by: sequence of
          (field names) = (FITS header keys) = (keys in Spectrum.more_headers)
    """

    def __init__(self, *args):
        XHelpDialog.__init__(self, *args)

        self.block = None
        self.group_by = None


        self.setWindowTitle("Group Spectra")

        self.labelHelpTopics.setText("&Available operations")

        self.help_data = a99.collect_doc(ex.blocks.gb, base_class=ex.GroupBlock)
        self.comboBox.addItems([x[0] for x in self.help_data])
        ###
        label = QLabel(a99.enc_name_descr("O&peration", "See help below"))
        edit = self.editFunction = QLineEdit("GB_UseNumPyFunc(np.mean)")
        label.setBuddy(edit)
        self.grid.addWidget(label, 0, 0)
        self.grid.addWidget(edit, 0, 1)
        ###
        label = QLabel(a99.enc_name_descr("&Group fieldnames", "Comma-separated without quotes"))
        edit = self.editGroupBy = QLineEdit("")
        label.setBuddy(edit)
        self.grid.addWidget(label, 1, 0)
        self.grid.addWidget(edit, 1, 1)

        a99.place_center(self, 800, 600)

    def accept(self):
        try:
            expr = str(self.editFunction.text())
            symbols_available = a99.module_to_dict(ex.blocks.gb)
            import numpy
            symbols_available["np"] = numpy
            block = eval(expr.strip(), {}, symbols_available)

            if not isinstance(block, ex.GroupBlock):
                raise RuntimeError("Expression does not evaluate to a valid Grouping Block")

            s_group_by = str(self.editGroupBy.text())
            group_by = [str(y).upper() for y in [x.strip() for x in s_group_by.split(",")] if len(y) > 0]

            self.block = block
            self.group_by = group_by

            return QDialog.accept(self)
        except Exception as e:
            self.add_log_error(a99.str_exc(e), True)
            return False




import f311.filetypes as ft
import datetime

__all__ = ["make_file_molecules"]


def make_file_molecules(mol_row, state_consts, lines, qgbd_calculator, sols_calculator):
    """

    Args:
        mol_row: MyDBRow object representing row from 'molecules' table
        state_consts: dict-like object with keys "omega_e", etc.
        lines: molecular lines data; type/format varies depending on type of data
        qgbd_calculator: callable(state_consts, v_lo) that can calculate "qv", "gv", "bv", "dv",
                         e.g., calc_qbdg_tio_like()
        sols_calculator: callable(state_consts, lines, qgbd_calculator) that returns a list of
                         SetOfLines, e.g., hitran_to_sols()

    Returns: (a FileMolecules object, a MolConversionLog object)
    """

    sols, log = sols_calculator(mol_row, state_consts, lines, qgbd_calculator)
    mol = _make_molecule(mol_row, sols)
    f = ft.FileMolecules()
    f.titm = "Created by ftpyfant.make_file_molecules() @ {}".format(datetime.datetime.now().isoformat())
    f.molecules = [mol]

    return f, log


def _make_molecule(mol_row, sols):
    """Assembles a Molecule object taking all necessary information

    Args:
        mol_row: MyDBRow object representing row from 'molecules' table
        mol_consts: dict-like object with keys "fe", "do", "am", "bm", "ua", "ub", "te"
        sol: ftpyfant.SetOfLines object

    **Note** Doesn't take take "fe", "do", etc. directly from the 'molecule' table because these
             values may have been customized by the user in widget WMolConst
    """

    mol = ft.Molecule()
    mol.description = "{name} ({formula})".format(**mol_row)
    mol.symbols = [mol_row["symbol_a"], mol_row["symbol_b"]]
    mol.fe = mol_row["fe"]
    mol.do = mol_row["do"]
    mol.mm = mol_row["am"] + mol_row["bm"]
    mol.am = mol_row["am"]
    mol.bm = mol_row["bm"]
    mol.ua = mol_row["ua"]
    mol.ub = mol_row["ub"]
    mol.te = mol_row["te"]
    mol.cro = mol_row["cro"]
    mol.s = mol_row["s"]

    mol.sol = sols

    return mol


__all__ = ["XRunnableManager"]


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from ..rm import RunnableManager
from threading import Lock
import time
import matplotlib.pyplot as plt
import a99
import f311.explorer as ex
import f311.filetypes as ft
from ... import pyfant as pf


COLORS = [QColor(255, 0, 0),
          QColor(192, 0, 0),
          QColor(96, 96, 96),
          QColor(128, 128, 128)]
STATUS_COLUMN_WIDTH = 730
WINDOW_WIDTH = 700
WINDOW_HEIGHT = 700
# # Defaults for checkable buttons (True: checked)
FLAG_SHOW_STATUS = True
FLAG_SHOW_RUNNABLES = True


class XRunnableManager(QMainWindow):
    """
    Thread manager window.

    Allows to monitor running runnables.
    """

    def __init__(self, parent=None, rm=None):
        QMainWindow.__init__(self, parent)
        assert isinstance(rm, RunnableManager)
        self.rm = rm
        # Whether to close matplotlib plots when window is closed
        self.flag_close_mpl_plots_on_close = True
        # Whether to show "close window to continue" when finished
        self.flag_close_message = True
        self.__num_finished_shown = 0  # outdated version of self.rm.num_finished
        self.__lock_table = Lock()
        self.__lfcs = 0  # labelFinished color state
        self.__explorer_form = None


        # # Window design

        # ## Toolbar

        b0 = self.pushButtonStatus = QPushButton("Show &status")
        b0.clicked.connect(self.on_status)
        b0.setCheckable(True)
        b0.setChecked(FLAG_SHOW_STATUS)
        b0.setToolTip("Turn status panel on/off")
        b1 = self.pushButtonTable = QPushButton("Show &runnables")
        b1.clicked.connect(self.on_table)
        b1.setCheckable(True)
        b1.setChecked(FLAG_SHOW_RUNNABLES)
        b1.setToolTip("Turn status panel on/off")
        b2 = self.pushButtonPause = QPushButton("&Pause")
        b2.clicked.connect(self.on_pause)
        b2.setCheckable(True)
        b2.setToolTip("Pause/Resume")
        b3 = self.pushButtonCancel = QPushButton("&Kill running")
        b3.setToolTip("Attempts to kill all running jobs")
        b3.clicked.connect(self.on_kill)
        b4 = self.pushButtonCancel = QPushButton("&Exit thread")
        b4.setToolTip("Stops runnables manager thread (irreversible)")
        b4.clicked.connect(self.on_exit)
        b5 = self.pushButtonRetryFailed = QPushButton("Retry &failed")
        b5.clicked.connect(self.on_retry_failed)
        b6 = self.pushButtonCollectErrors = QPushButton("&Collect errors")
        b6.setToolTip("Searches for errors in log files and reports these errors in a new window.")
        b6.clicked.connect(self.on_collect_errors)
        # todo + note: assuming that there is no interest in adding new tasks later
        l = self.labelEnd = QLabel()
        l.setVisible(False)

        s = self.spacer0 = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)

        bb = [b0, b1, b2, b3, b4, b5, b6, l]
        l0 = self.layoutToolbar = QHBoxLayout()
        for b in bb:
            l0.addWidget(b)
        l0.addItem(s)

        a = self.widgetToolbar = QWidget()
        a.setLayout(l0)
        a.setFixedHeight(40)

        # ## Status widget

        # ### Left column
        w0 = self.labelStatusFinished = QLabel()
        w1 = self.labelStatusFailed = QLabel()
        w2 = self.labelStatusElla = QLabel()
        w3 = self.labelStatusTPR = QLabel()
        w4 = self.labelStatusTotal = QLabel()
        w5 = self.labelStatusRema = QLabel()
        ww = [("Num finished", w0),
              ("Num failed", w1),
              ("Time ellapsed", w2),
              ("Time total estimate", w4),
              ("Time remaining estimate", w5),
              ("Average time per runnable", w3)]
        l = self.layoutStatus0 = QFormLayout()
        for caption, w in ww:
            w.setStyleSheet("QLabel {font-weight: bold}")
            l.addRow(caption, w)

        # ### Right column
        w0 = self.checkBox_paused = QCheckBox()
        w1 = self.checkBox_cancelled = QCheckBox()
        w2 = self.checkBox_finished = QCheckBox()
        w3 = self.checkBox_failed = QCheckBox()
        w4 = self.checkBox_exited = QCheckBox()
        l = self.layoutStatus1 = QFormLayout()
        ww = [("Paused", w0),
              ("Cancelled", w1),
              ("Finished", w2),
              ("Failed", w3),
              ("Exited", w4)]
        for caption, w in ww:
            w.setChecked(False)
            w.setEnabled(False)
            l.addRow(caption, w)

        l = QHBoxLayout()
        l.addLayout(self.layoutStatus0)
        l.addLayout(self.layoutStatus1)
        x = self.widgetStatus = QWidget()
        x.setVisible(b0.isChecked())
        x.setLayout(l)


        # ## Table Widget

        a = self.tableWidget = QTableWidget()
        a.setToolTip("Double-click to open session directory")
        a.setSelectionMode(QAbstractItemView.SingleSelection)
        a.setAlternatingRowColors(True)
        a.setEditTriggers(QTableWidget.NoEditTriggers)
        a.setSelectionBehavior(QTableWidget.SelectRows)
        a.cellDoubleClicked.connect(self.on_tableWidget_cellDoubleClicked)
        # a.setFont(MONO_FONT)
        a.setVisible(self.pushButtonTable.isChecked())
        a.installEventFilter(self)

        # ## Mounts central widget

        l2 = self.layoutCentral = QVBoxLayout()
        ww = [self.widgetToolbar, self.widgetStatus, self.tableWidget]
        for w in ww:
            l2.addWidget(w)
        # l2.addWidget(self.widgetPlot)
        a99.set_margin(l2, 0)
        a = self.centralWidget = QWidget()
        a.setLayout(l2)
        a.setFont(a99.MONO_FONT)
        self.setCentralWidget(self.centralWidget)



        # # # Status bar
        # self.label_n = QLabel()
        # self.label_t = QLabel()
        # sb = self.statusBar()
        # sb.insertWidget(0, self.label_n, 0)
        # sb.insertWidget(1, self.label_t, 0)


        # # Final adjustments
        self.setGeometry(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT)
        a99.place_center(self)
        self.setWindowTitle("Runnables Manager")

        # # Non-visual stuff

        # ## Wiring of signals to trigger status update
        t = self.timerUpdate = QTimer()
        t.setInterval(1000)  # miliseconds
        signals = [t.timeout, self.rm.runnable_changed, self.rm.finished]
        self.changed_proxy = a99.SignalProxy(signals,
        delay=0, rateLimit=1, slot=self.__update, flag_connect=False)
        self.rm.runnable_added.connect(self.__populate, Qt.QueuedConnection)

        # ## Timer to flick the finish indicator
        t = self.timerFinished = QTimer()
        t.setInterval(200)  # miliseconds
        t.timeout.connect(self.on_timerFinished_timeout)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def showEvent(self, event):
        self.__populate()
        self.changed_proxy.connect_all()
        # self.rm.runnable_added.connect(self.on_tm_thread_added, Qt.QueuedConnection)
        self.timerUpdate.start()

    def closeEvent(self, event):
        if not self.rm.flag_finished:
            r = QMessageBox.question(self, "Close window", "Not finished yet. Are you sure?",
             QMessageBox.Yes|QMessageBox.No, QMessageBox.Yes)
            if r != QMessageBox.Yes:
                event.ignore()
                return
        self.changed_proxy.disconnect_all()
        # self.rm.runnable_added.disconnect(self.on_tm_thread_added)
        self.timerUpdate.stop()
        self.timerFinished.stop()

        if self.flag_close_mpl_plots_on_close:
            plt.close("all")

    def eventFilter(self, obj, event):
        if obj == self.tableWidget:
            return a99.check_return_space(event, self.on_tableWidget_cellDoubleClicked)
        return False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    # def on_tm_thread_changed(self):
    #     print "on_tm_thread_changed"
    #     self._update()

    # def on_tm_thread_added(self):
    #     self._populate()

    def on_pause(self):
        if self.pushButtonPause.isChecked():
            self.rm.pause()
        else:
            self.rm.resume()

    def on_status(self):
        self.widgetStatus.setVisible(self.pushButtonStatus.isChecked())

    def on_table(self):
        self.tableWidget.setVisible(self.pushButtonTable.isChecked())

    def on_exit(self):
        self.rm.cancel()

    def on_kill(self):
        self.rm.kill_runnables()

    def on_retry_failed(self):
        try:
            self.rm.retry_failed()
        except Exception as e:
            MSG = "Could not retry failed"
            a99.get_python_logger().exception(MSG)
            a99.show_error("%s: %s" % (MSG, str(e)))

    def on_collect_errors(self):
        try:
            k = a99.ErrorCollector()
            k.collect_errors(".")
            w = ex.XHTML(self, k.get_html(), "Errors in '.' and subdirectories")
            w.show()
        except Exception as e:
            MSG = "Could not collect errors"
            a99.get_python_logger().exception(MSG)
            a99.show_error("%s: %s" % (MSG, str(e)))

    # def on_timer_timeout(self):
    #     print "on_timer_timeout"
    #     self._update()

    def on_timerFinished_timeout(self):
        self.__set_labelFinished_color(COLORS[self.__lfcs])
        self.__lfcs += 1
        if self.__lfcs == len(COLORS):
            self.__lfcs = 0

    def on_tableWidget_cellDoubleClicked(self, row=0, col=0):
        try:
            self.__explore_directory()
        except Exception as e:
            MSG = "Could explore directory"
            a99.get_python_logger().exception(MSG)
            a99.show_error("%s: %s" % (MSG, str(e)))


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Gear

    def __populate(self):
        """Clears and rebuilds table widget."""
        with self.__lock_table:
            a = self.tableWidget
            self.runnables = runnables = self.rm.get_runnables_copy()
            a.clear()
            self.__num_finished_shown = 0
            a.setAlternatingRowColors(True)
            a.setRowCount(len(runnables))
            a.setColumnCount(2)
            a.setHorizontalHeaderLabels(["Session directory", "Status"])
            for i, runnable in enumerate(runnables):
                title = runnable.sid.dir
                if title is None:
                    title = '...'
                item = QTableWidgetItem(title)
                a.setItem(i, 0, item)
                status = str(runnable.get_status())
                item = QTableWidgetItem("?" if status is None else str(status))
                a.setItem(i, 1, item)
            a.resizeColumnsToContents()
            a.setColumnWidth(1, STATUS_COLUMN_WIDTH)
            self.__update_status()

    def __update(self):
        """Updates second column of table widget."""
        # a99.get_python_logger().info("UPDATE UPDATE UPDATE UPDATE UPDATE ")
        t = time.time()
        self.__update_status()
        self.__update_table()
        # a99.get_python_logger().info("&&&&&&&&&&&&&&& time to update: %g" % (time.time()-t,))

    def __update_table(self):
        with self.__lock_table:
            a = self.tableWidget
            runnables = self.runnables
            nf = self.rm.num_finished  # grabs this before last table update,
                                       # so that it never skips a row update
            mt = self.rm.max_simultaneous
            for i in range(max(0, self.__num_finished_shown-mt+1),
                            min(len(runnables), self.rm.num_finished+mt)):
                runnable = runnables[i]

                item = a.item(i, 0)
                title = runnable.sid.dir
                if title is None:
                    title = "..."
                item.setText(title)

                item = a.item(i, 1)
                status = runnable.get_status()
                item.setText("?" if status is None else str(status))

            if nf != self.__num_finished_shown:
                a.setCurrentCell(self.rm.num_finished+self.rm.max_simultaneous-1, 0)
                self.__num_finished_shown = nf

    def __update_status(self):
        """Updates everything except the table widget."""
        flag_todo = not (self.rm.flag_finished or self.rm.flag_exited)
        self.pushButtonCancel.setEnabled(flag_todo)
        self.pushButtonPause.setEnabled(flag_todo)
        if not flag_todo and self.flag_close_message:
            s = "Please close window to continue"
            self.labelEnd.setText(s)
            self.labelEnd.setVisible(True)
            self.timerFinished.start()
        if self.pushButtonStatus.isChecked():
            self.labelStatusFinished.setText("%d/%d" %
             (self.rm.num_finished, self.rm.num_runnables))
            self.labelStatusFailed.setText(str(self.rm.num_failed))
            ella, tot, rema = self.rm.get_times()
            self.labelStatusElla.setText(a99.seconds2str(ella))
            self.labelStatusTPR.setText(a99.seconds2str(self.rm.time_per_runnable))
            self.labelStatusTotal.setText(a99.seconds2str(tot))
            self.labelStatusRema.setText(a99.seconds2str(rema))

            self.checkBox_paused.setChecked(self.rm.flag_paused)
            self.checkBox_cancelled.setChecked(self.rm.flag_cancelled)
            self.checkBox_finished.setChecked(self.rm.flag_finished)
            self.checkBox_failed.setChecked(self.rm.flag_failed)
            self.checkBox_exited.setChecked(self.rm.flag_exited)

    def __set_labelFinished_color(self, color):
        self.labelEnd.setStyleSheet("QLabel {color: "+color.name()+";}")

    def __explore_directory(self):
        runnable = self.runnables[self.tableWidget.currentRow()]
        dir_ = runnable.sid.dir
        if not self.__explorer_form:
            f = self.__explorer_form = ex.XExplorer(self, dir_)
            f.flag_close_mpl_plots_on_close = False
        else:
            self.__explorer_form.set_dir(dir_)
        self.__explorer_form.show()
        self.__explorer_form.raise_()
        self.__explorer_form.activateWindow()




"""
Rainbow colors

Example:

>>> for color in rainbow_colors:
...     color
Color('Violet', [ 0.54509804  0.          1.        ], 4000.0, 3775.0, 4225.0)
Color('Indigo', [ 0.29411765  0.          0.50980392], 4450.0, 4225.0, 4600.0)
Color('Blue', [ 0.  0.  1.], 4750.0, 4600.0, 4925.0)
Color('Green', [ 0.  1.  0.], 5100.0, 4925.0, 5400.0)
Color('Yellow', [ 1.  1.  0.], 5700.0, 5400.0, 5800.0)
Color('Orange', [ 1.          0.49803922  0.        ], 5900.0, 5800.0, 6200.0)
Color('Red', [ 1.  0.  0.], 6500.0, 6200.0, 6800.0)

Example:

>>> import matplotlib.pyplot as plt
>>> for color in rainbow_colors:
...     _ = plt.fill_between([color.l0, color.lf], [1., 1.], color=color.rgb, label=color.name)
>>> _ = plt.legend(loc=0)
>>> _ = plt.xlabel('Wavelength (angstrom)')
>>> _ = plt.tight_layout()
>>> _ = plt.show()

"""


__all__ = ["Color", "rainbow_colors", "ncolors", "spectrum_to_rgb"]


from a99 import AttrsPart
import numpy as np


class Color(AttrsPart):
    """Definition of a color: name, RGB code, wavelength range"""

    attrs = ["name", "rgb", "clambda", "l0", "lf"]

    def __init__(self, name, rgb, clambda, l0=-1, lf=-1):
        AttrsPart.__init__(self)
        # Name of color
        self.name = name
        # (red, green, blue) tuple (numbers in [0, 1] interval)
        self.rgb = rgb
        # Central wavelength (angstrom)
        self.clambda = clambda
        # Initial wavelength (angstrom)
        self.l0 = l0
        # Final wavelength (angstrom)
        self.lf = lf

    def __repr__(self):
        return "Color('{}', {}, {}, {}, {})".format(self.name, self.rgb, self.clambda, self.l0, self.lf)


#: Rainbow colors
rainbow_colors = [Color("Violet", [139, 0, 255], 4000),
                  Color("Indigo", [75, 0, 130], 4450),
                  Color("Blue", [0, 0, 255], 4750),
                  Color("Green", [0, 255, 0], 5100),
                  Color("Yellow", [255, 255, 0], 5700),
                  Color("Orange", [255, 127, 0], 5900),
                  Color("Red", [255, 0, 0], 6500),
                  ]

# Calculates l0, lf
# rainbow_colors[0].l0 = 0.
# rainbow_colors[-1].lf = float("inf")
for c in rainbow_colors:
    c.clambda = float(c.clambda)
ncolors = len(rainbow_colors)
for i in range(1, ncolors):
    cprev, cnow = rainbow_colors[i-1], rainbow_colors[i]
    avg = (cprev.clambda + cnow.clambda) / 2
    cprev.lf = avg
    cnow.l0 = avg

    if i == 1:
        cprev.l0 = 2*cprev.clambda-cprev.lf
    if i == ncolors-1:
        cnow.lf = 2*cnow.clambda-cnow.l0
# converts RGB from [0, 255] to [0, 1.] interval
for c in rainbow_colors:
    c.rgb = np.array([float(x)/255 for x in c.rgb])


def spectrum_to_rgb(sp, visible_range=None, method=0):
    """Takes weighted average of rainbow colors RGB's

    Args:
        sp: Spectrum instance
        visible_range=None: if passed, affine-transforms the rainbow colors
        method:

            - ``0``: rainbow colors
            - ``1``: RGB (alternative method using red, green, blue only instead of full range of rainbow colors)

    Returns:
        3-element sequence where each element is in [0, 1] range

    Example:

    >>> import f311.physics as ph
    >>> vega = ph.get_vega_spectrum()
    >>> ph.spectrum_to_rgb(vega)
    array([  5.45057920e-01,   1.61124778e-05,   9.99930126e-01])
    """

    if visible_range is None:
        visible_range = min(sp.x), max(sp.x)

    if len(visible_range) < 2:
        raise RuntimeError("Invalid visible range: {0!s}".format(visible_range))
    if visible_range[1] <= visible_range[0]:
        raise RuntimeError(
            "Second element of visible range ({0!s}) must be greater than first element".format(
                visible_range))

    if method == 1:
        # new color system splitting visible range in three
        dl = float(visible_range[1] - visible_range[0]) / 3
        ranges = np.array(
            [[visible_range[0] + i * dl, visible_range[0] + (i + 1) * dl] for i in range(3)])
        colors = np.array([(0., 0., 1.), (0., 1., 0.), (1., 0., 0.)])  # blue, green, red

        tot_area, tot_sum = 0., np.zeros(3)
        for color, range_ in zip(colors, ranges):
            b = np.logical_and(sp.x >= range_[0], sp.x <= range_[1])
            area = np.sum(sp.y[b])
            tot_area += area
            tot_sum += color * area
        if tot_area == 0.:
            tot_area = 1.
        ret = tot_sum / tot_area
        return ret

    elif method == 0:
        tot_area, tot_sum = 0., np.zeros(3)

        def ftrans(x):
            return x

        if visible_range:
            ll0 = rainbow_colors[0].l0
            llf = rainbow_colors[-1].lf
            ftrans = lambda lold: visible_range[0] + (visible_range[1] - visible_range[0]) / (
            llf - ll0) * (lold - ll0)

        for color in rainbow_colors:
            b = np.logical_and(sp.x >= ftrans(color.l0), sp.x <= ftrans(color.lf))
            area = np.sum(sp.y[b])
            tot_area += area
            tot_sum += color.rgb * area
        if tot_area == 0.:
            tot_area = 1.
        ret = tot_sum / tot_area
        return ret
    else:
        raise RuntimeError("Unknown method: {0!s}".format(method))



"""
NIST Downloader
"""
from robobrowser import RoboBrowser
import bs4
import re


__all__ = ["get_nist_webbook_constants"]


_conv_sup = {"1": "\u2071",
             "2": "\u00b2",
             "3": "\u00b3",
             "4": "\u2074",
             "+": "\u207a",
             "-": "\u207b"}


_conv_img = {"larrow": "←",
             "rarrow": "→",
             "lrarrow": "↔"}


def _html_formula_to_unicode(tag):
    """
    Returns a string from a bs4.element.Tag representing a formula in NIST web book pages.

    Some HTML elements are converted to unicode characters
    """
    parts = []
    for item in tag.children:
        if isinstance(item, bs4.element.NavigableString):
            parts.append(str(item))
        elif item.name == "sub":
            parts.append("_" + item.text)
        elif item.name == "sup":
            parts.append(_conv_sup.get(item.text, "^" + item.text))
        elif item.name == "img":
            # parts.append("\\"+item.attrs["alt"])
            alt = item.attrs["alt"]
            parts.append(_conv_img.get(alt, alt))
    return ("".join(parts)).strip().replace("  ", " ").replace("  ", " ")


def _floatify(str_):
    """Tries to convert to float, stripping brackets if it is the case"""
    try:
        ret = float(str_)
    except:
        gg = re.match("[\[(]*([0-9.-]+)", str_)
        if gg is None:
            ret = str_
        else:
            ret = float(gg.groups()[0])
    return ret

def _nonify(x):
    """Returns None if len(x) == 0, else x"""
    if isinstance(x, str) and len(x.strip()) == 0:
        return None
    return x


# Callables to be used in different conversion situations
f_header = _html_formula_to_unicode
f_state = _html_formula_to_unicode
f_float_or_none = lambda x: _nonify(_floatify(_html_formula_to_unicode(x)))
f_str_or_none = lambda x: _nonify(_html_formula_to_unicode(x))
functions = [f_state]+[f_float_or_none]*10+[f_str_or_none, f_float_or_none]



def get_nist_webbook_constants(formula):
    """
    Navigates through NIST webbook pages to retrieve a table of molecular constants

    Args:
        formula: example: "OH"

    Returns: tuple: table (list of lists), header (list of strings), name of molecule
    """
    browser = RoboBrowser(history=True, parser="lxml")
    browser.open("http://webbook.nist.gov/chemistry/form-ser.html")

    # # Page
    form = browser.get_form()
    form["Formula"].value = formula
    browser.submit_form(form)

    # # Molecule park page
    text = browser.find(text="Constants of diatomic molecules")

    if text is None:
        # # Probably fell in this page showing possible choices
        link0 = browser.find("a")  # follows first link
        if not link0:
            raise RuntimeError("Search for '{}' gave no results".format(formula))
        browser.follow_link(link0)

        # # Molecule park page
        text = browser.find(text="Constants of diatomic molecules")
        link_ctes = text.parent
        browser.follow_link(link_ctes)
    else:
        link_ctes = text.parent
        browser.follow_link(link_ctes)

    # # Page

    title = browser.find("title").text

    table = browser.find("table", class_="small data")

    rows = table.find_all("tr")
    n = 0
    header = None
    data = []  # list of lists
    for row in rows:
        cols = row.find_all("td")
        flag_header = False
        if len(cols) == 0:
            if n > 0:
                continue
            # If no columns found, we assume it is the header row
            cols = row.find_all("th")
            header = [f_header(ele) for ele in cols]
        elif len(cols) == 13:
            row = [function(ele) for function, ele in zip(functions, cols)]
            data.append(row)
            n += 1

    # print(plain)
    #
    # print(rows_table)

    return data, header, title



__all__ = ["FullCube"]


import a99
import numpy as np
from astropy.io import fits


@a99.froze_it
class FullCube(a99.AttrsPart):
    """
    X-Y-wavelength cube

    Based on IDL source file chris_J4000.pro

    Data is stored primarily in self.hdu, with a few other relevant attributes

    Notes on FITS HDU:
      - data is stored in reverse dimensions, i.e., (z, y, x)
      - CDELT1, CRVAL1 refer to x
      - CDELT2, CRVAL2 refer to y
      - CDELT3, CRVAL3 refer to z
    """

    attrs = ["hdu"]

    @property
    def flag_created(self):
        """Whether the data cube has already been created"""
        return self.hdu is not None

    @property
    def flag_wavelengthed(self):
        """Whether wavelength information is already present """
        return self.flag_created and self.hdu.header["CDELT3"] != -1

    def __init__(self, hdu=None):
        a99.AttrsPart.__init__(self)
        self.hdu = None  # PyFITS HDU object
        self.wavelength = None  # the wavelength axis (angstrom) (shared among all spectra in the cube)
        self.filename = None

        if hdu is not None:
            self.from_hdu(hdu)

    def from_hdu(self, hdu):
        """
        Sets self.hdu and with HDU object passed. Warning: may change HDU object.

        Default HDU headers:

            - CDELT1 defaults to 1
            - CDELT2 defaults to CDELT1
            - CRDELT3 defaults to 1
            - CRVAL3 defaults to 0
        """
        assert isinstance(hdu, fits.PrimaryHDU)

        # ensures all required headers are present
        keys = ["NAXIS1", "NAXIS2", "NAXIS3"]
        for key in keys:
            assert key in hdu.header, 'Key "%s" not found in headers' % key

        if not "CDELT3" in hdu.header:
            a99.get_python_logger().warning("HDU lacks header 'CDELT3', assumed 1.0")
            hdu.header["CDELT3"] = 1.
        if not "CRVAL3" in hdu.header:
            a99.get_python_logger().warning("HDU lacks header 'CRVAL3', assumed 0.0")
            hdu.header["CRVAL3"] = 0.
        if not "CDELT1" in hdu.header:
            a99.get_python_logger().warning("HDU lacks header 'CDELT1', assumed 1.0")
            hdu.header["CDELT1"] = 1.0
        if not "CDELT2" in hdu.header:
            a99.get_python_logger().warning("HDU lacks header 'CDELT2', assumed value in 'CDELT1'")
            hdu.header["CDELT2"] = hdu.header["CDELT1"]
        if not "HRFACTOR" in hdu.header:
            hdu.header["HRFACTOR"] = 1.

        l0 = hdu.header["CRVAL3"]  # I think WC was adding this but changed ... +hdu.header["CDELT3"]
        delta_lambda = hdu.header["CDELT3"]
        nlambda = hdu.header["NAXIS3"]
        self.hdu = hdu
        self.set_wavelength(np.array([l0 + k * delta_lambda for k in range(0, nlambda)]))

    def __len__(self):
        raise NotImplementedError()

    def __repr__(self):
        return "Please implement FullCube.__repr__()"

    def create1(self, R, dims, hr_pix_size, hrfactor):
        """Creates FITS HDU, including the cube full with zeros

          dims: (nlambda, height, width)
        """
        cube = np.zeros(dims)
        hdu = fits.PrimaryHDU()
        hdu.header["CDELT1"] = hr_pix_size
        hdu.header["CDELT2"] = hr_pix_size
        hdu.header["CDELT3"] = -1.0  # this will become known then paint() is called for the first time
        hdu.header["CRVAL3"] = -1.0  # "
        hdu.header["HRFACTOR"] = hrfactor
        hdu.header["R"] = R
        hdu.data = cube
        self.hdu = hdu

    def get_spectrum(self, x, y):
        """
        Returns spectrum at coordinate x, y (copied vectors)

        **Note** coordinate (x=0, y=0) corresponds to lower left pixel of cube cross-section
        """
        import f311.filetypes as ft
        assert self.flag_wavelengthed

        sp = ft.Spectrum()
        sp.x = np.copy(self.wavelength)
        sp.y = np.copy(self.hdu.data[:, y, x])
        sp.more_headers["PIXEL-X"] = x  # why not
        sp.more_headers["PIXEL-Y"] = y
        return sp

    def set_wavelength(self, w):
        delta_lambda = w[1] - w[0]
        self.hdu.header["CDELT3"] = delta_lambda
        # dunno why the initial lambda must be one delta lambda lower than the initial lambda in the spectrum, but
        # this is how it was in the original
        self.hdu.header["CRVAL3"] = w[0] - delta_lambda
        self.wavelength = w



"""
"""


import f311.physics as ph
import f311.filetypes as ft
from .. import convmol as cm
import a99
from .convlog import *
from collections import OrderedDict
import sys

__all__ = ["kurucz_to_sols"]



def kurucz_to_sols(mol_row, state_row, fileobj, qgbd_calculator, flag_hlf=False, flag_normhlf=False,
                   flag_fcf=False):
    """
    Converts Kurucz molecular lines data to PFANT "sets of lines"

    Args:
        mol_row: dict-like,
                 molecule-wide constants,
                 keys: same as as field names in 'moldb:molecule' table
        state_row: dict-like,
                   state-wide constants,
                   keys: same as field names in 'moldb:state' table
        fileobj: FileVald3 instance with only one species
        qgbd_calculator: callable that can calculate "qv", "gv", "bv", "dv",
                         e.g., calc_qbdg_tio_like()
        flag_hlf: Whether to calculate the gf's using Honl-London factors or
                  use Kurucz's loggf instead

        flag_normhlf: Whether to multiply calculated gf's by normalization factor


    Returns: (a list of ftpyfant.SetOfLines objects, a MolConversionLog object)
    """

    def append_error(msg):
        log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))

    # C     BAND (v',v'')=(VL,V2L)
    # C     WN: vacuum wavenumber   WL : air wavelength
    # C     J2L: lower state angular momentum number
    # C     iso: isotope/ 26: 12C16O, 36: 13C16O, 27: 12C17O, 28: 12C18O
    # C     ramo: branch, for P: ID=1, for R: ID=2
    # C     name: file name/ coisovLv2L
    # C     HL: Honl-London factor
    # C     FR: oscillator strength

    if not isinstance(fileobj, ft.FileKuruczMolecule):
        raise TypeError("Invalid type for argument 'fileobj': {}".format(type(fileobj).__name__))

    lines = fileobj.lines
    n = len(lines)

    S = mol_row["s"]
    DELTAK = mol_row["cro"]
    fe = mol_row["fe"]

    # TODO of course this hard-wire needs change; now just a text for OH A2Sigma-X2Pi
    LAML = 0  # Sigma
    LAM2L = 1 # Pi

    if flag_hlf:
        formulas = ph.doublet.get_honllondon_formulas(LAML, LAM2L)
    sols = OrderedDict()  # one item per (vl, v2l) pair
    log = MolConversionLog(n)


    # This factor allows to reproduce the Hônl-London factors in `moleculagrade.dat` for OH blue,
    # first set-of-lines
    scale_factor = 730.485807466/2

    for i, line in enumerate(lines):
        assert isinstance(line, ft.KuruczMolLine)
        # TODO is it spin 2l?
        branch = ph.doublet.quanta_to_branch(line.Jl, line.J2l, line.spin2l)
        try:
            # TODO take this outta here, for now just to filter to match others
            fcf = cm.get_fcf_oh(line.vl, line.v2l)

            wl = line.lambda_

            # Normaliza = 1/((2.0*line.J2l+1)*(2.0*S+1)*(2.0-DELTAK))

            if flag_normhlf:
                # k = 2 / ((2.0*line.J2l+1)*(2.0*S+1)*(2.0-DELTAK))
                # k = 2 / ((2.0*line.J2l+1))
                k = 2/ ((2*S+1) * (2*line.J2l+1) * (2-DELTAK))
                # k = (2.0*line.J2l+1)
                # k = (2*S+1) * (2*line.J2l+1) * (2-DELTAK)
            else:
                k = 1

#            k *= fe

            if flag_hlf:
                hlf = formulas[branch](line.J2l)
                gf_pfant = hlf*k

            else:
                # Normaliza = scale_factor * k
                gf_pfant = k*10**line.loggf

            # # TODO ask BLB for references on these Franck-Condon Factors & try to generalize (these are for OH only I think)
            # x = line.J2l * (line.J2l + 1)
            # if branch[0] == "P":
            #     fcf = 3.651E-2 * (1 + 4.309E-6 * x + 1.86E-10 * (x ** 2)) ** 2
            # elif branch[0] == "Q":
            #     fcf = 3.674E-2 * (1 + 6.634E-6 * x + 1.34E-10 * (x ** 2)) ** 2
            # elif branch[0] == "R":
            #     fcf = 3.698E-2 * (1 + 1.101E-5 * x + 7.77E-11 * (x ** 2)) ** 2
            # else:
            #     raise RuntimeError("Shouldn't fall in here (branch ie neither P/Q/R)")

 #           fcf = 1.3201e-1

            if flag_fcf:
                fcf = cm.get_fcf_oh(line.vl, line.v2l)

                gf_pfant *= fcf

                # if branch[0] == "P":
                #     gf_pfant *= 3.651E-2 * (1 + 4.309E-6 * x + 1.86E-10 * (x ** 2)) ** 2
                # elif branch[0] == "Q":
                #     gf_pfant *= 3.674E-2 * (1 + 6.634E-6 * x + 1.34E-10 * (x ** 2)) ** 2
                # elif branch[0] == "R":
                #     gf_pfant *= 3.698E-2 * (1 + 1.101E-5 * x + 7.77E-11 * (x ** 2)) ** 2
                # else:
                #     raise RuntimeError("Shouldn't fall in here (branch ie neither P/Q/R)")

            J2l_pfant = line.J2l
        except Exception as e:
            # if isinstance(e, ZeroDivisionError):
            #     print("OOOOOOOOOOOOOOOOOOOOOO")
            #     print(str(line))
            #     sys.exit()

            msg = "#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), a99.str_exc(e))
            log.errors.append(msg)
            a99.get_python_logger().exception(msg)
            continue

        sol_key = "%3d%3d" % (line.vl, line.v2l)  # (v', v'') transition (v_sup, v_inf)
        if sol_key not in sols:
            qgbd = qgbd_calculator(state_row, line.v2l)
            qqv = qgbd["qv"]
            ggv = qgbd["gv"]
            bbv = qgbd["bv"]
            ddv = qgbd["dv"]
            sols[sol_key] = ft.SetOfLines(line.vl, line.v2l, qqv, ggv, bbv, ddv, 1.)

        sol = sols[sol_key]
        sol.append_line(wl, gf_pfant, J2l_pfant, branch)

    return (list(sols.values()), log)



__all__ = ["XFileSparseCube"]

import collections
import copy
import matplotlib.pyplot as plt
from pylab import MaxNLocator
import numbers
import numpy as np
import os
import os.path
from itertools import product, combinations, cycle
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_WFileSparseCube import *
import a99
from .... import explorer as ex
import f311.filetypes as ft


class XFileSparseCube(a99.XFileMainWindow):
    def __init__(self, parent=None, fileobj=None):
        a99.XFileMainWindow.__init__(self, parent)

        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        self.setWindowTitle(a99.get_window_title("Data Cube Editor"))


        # # Synchronized sequences
        _VVV = ft.FileSparseCube.description
        self.tab_texts[0] =  "FileSparseCube editor (Alt+&1)"
        self.tabWidget.setTabText(0, self.tab_texts[0])
        self.save_as_texts[0] = "Save %s as..." % _VVV
        self.open_texts[0] = "Load %s" % _VVV
        self.clss[0] = ft.FileSparseCube
        self.clsss[0] = (ft.FileSparseCube, ft.FileFullCube)  # file types that can be opened
        self.wilds[0] = "*.fits"

        lv = keep_ref(QVBoxLayout(self.gotting))
        ce = self.ce = WFileSparseCube(self)
        lv.addWidget(ce)
        ce.edited.connect(self.on_tab0_file_edited)
        self.editors[0] = ce

        # # # Loads default file by default ... SQN
        # if os.path.isfile(FileSparseCube.default_filename):
        #     f = ft.FileSparseCube()
        #     f.load()
        #     self.ce.load(f)

        if fileobj is not None:
            self.load(fileobj)


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Qt override

    def closeEvent(self, event):
        flag_exit, ff = True, []
        for ed, flag_changed in zip(self.editors, self.flags_changed):
            if ed and ed.f and flag_changed:
                ff.append(ed.f.description)

        if len(ff) > 0:
            s = "Unsaved changes\n  -"+("\n  -".join(ff))+"\n\nAre you sure you want to exit?"
            flag_exit = a99.are_you_sure(True, event, self, "Unsaved changes", s)
        if flag_exit:
            plt.close("all")

    def keyPressEvent(self, evt):
        incr = 0
        if evt.modifiers() == Qt.ControlModifier:
            n = self.tabWidget.count()
            if evt.key() in [Qt.Key_PageUp, Qt.Key_Backtab]:
                incr = -1
            elif evt.key() in [Qt.Key_PageDown, Qt.Key_Tab]:
                incr = 1
            if incr != 0:
                new_index = self._get_tab_index() + incr
                if new_index < 0:
                    new_index = n-1
                elif new_index >= n:
                    new_index = 0
                self.tabWidget.setCurrentIndex(new_index)

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for Qt library signals

    # def on_show_rm(self):
    #     if self._manager_form:
    #         self._manager_form.show()
    #         self._manager_form.raise_()
    #         self._manager_form.activateWindow()


    def on_tab0_file_edited(self):
        self._on_edited()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Protected methods to be overriden or used by descendant classes

    def _filter_on_load(self, f):
        """Converts from FileFullCube to FileSparseCube format, if necessary"""
        f1 = None
        if isinstance(f, ft.FileFullCube):
            f1 = ft.FileSparseCube()
            f1.sparsecube.from_full_cube(f.wcube)
        if f1:
            f1.filename = a99.add_bits_to_path(f.filename, "imported-from-",
                                           os.path.splitext(ft.FileSparseCube.default_filename)[1])
            f = f1
        return f



""""
List of spectra sharing same wavenumber axis. Uses FITS format
"""

__all__ = ["FileSpectrumList"]


from . import SpectrumList
from astropy.io import fits
import a99
from .. import DataFile


@a99.froze_it
class FileSpectrumList(DataFile):
    """FITS Spectrum List"""
    attrs = ['splist']
    description = "Spectrum List"
    default_filename = "default.splist"
    flag_txt = False

    def __init__(self):
        DataFile.__init__(self)
        self.splist = SpectrumList()

    def _do_load(self, filename):
        fits_obj = fits.open(filename)
        self.splist = SpectrumList()
        self.splist.from_hdulist(fits_obj)
        self.filename = filename

    def _do_save_as(self, filename):
        hdul = self.splist.to_hdulist()
        a99.overwrite_fits(hdul, filename)

    def init_default(self):
        # Already created OK
        pass



from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
# from a_WState import WState
# import moldb as db
from .a_WFileMolDB import *
import os
import f311.filetypes as ft


__all__ = ["XFileMolDB"]


class XFileMolDB(a99.XFileMainWindow):
    def __init__(self, parent=None, fileobj=None):
        a99.XFileMainWindow.__init__(self, parent)

        # # Synchronized sequences
        _VVV = ft.FileMolDB.description
        self.tab_texts[0] =  "{} (Alt+&1)".format(_VVV)
        self.tabWidget.setTabText(0, self.tab_texts[0])
        self.save_as_texts[0] = "Save %s as..." % _VVV
        self.open_texts[0] = "Load %s" % _VVV
        self.clss[0] = ft.FileMolDB
        self.clsss[0] = (ft.FileMolDB,)
        self.wilds[0] = "*.sqlite"


        lv = self.keep_ref(QVBoxLayout(self.gotting))
        me = self.moldb_editor = WFileMolDB(self)
        lv.addWidget(me)
        me.edited.connect(self._on_edited)
        self.editors[0] = me


        if fileobj is not None:
            self.load(fileobj)

    def wants_auto(self):
        idx = self.w_source.index
        filename = None
        if idx == 0:
            lines = self.w_hitran.data
            if lines:
                filename = "{}.dat".format(lines["header"]["table_name"])

        if filename is None:
            # Default
            filename = a99.new_filename("mol", "dat")

        self.w_out.value = filename


    def on_fill_missing(self):
        self.w_mol.None_to_zero()
        self.w_state.None_to_zero()

    # def mol_id_changed(self):
    #     id_ = self.w_mol.w_mol.id
    #     row = self.w_mol.w_mol.row
    #     self.w_state.set_id_molecule(id_)
    #     s = "States (no molecule selected)" if not row else "Select a State for molecule '{}'".format(row["formula"])
    #     self.title_state.setText(a99.format_title0(s))


    # # Override
    #   ========







def quanta_to_branch(Jl, J2l):
    """
    Singlet only has branches P/Q/R

    Args:
        Jl: J upper or J'
        J2l: J lower or  J''

    Returns: branch letter: "P"/"Q"/"R"
    """
    if Jl > J2l:
        return "R"
    elif Jl == J2l:
        return "Q"
    return "P"


#!/usr/bin/env python3

"""
Prints table of diatomic molecular constants

If ID is specified, prints data for single molecule;
otherwise, prints full table
"""

import argparse
import logging
import a99


a99.logging_level = logging.INFO

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('ID', type=str, help='HITRAN Molecule ID',
                        default='(all)', nargs='?')
    args = parser.parse_args()

    kwargs = {}
    if not args.ID == "(all)":
        kwargs["molecule.ID"] = args.ID

    pf.hitrandb.print_isotopologues(**kwargs)


__all__ = ["WSpectrumCollection"]

import copy
import os
import os.path
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from .a_XScaleSpectrum import *
from astropy import units as u
import a99
from .... import explorer as ex
import f311.filetypes as ft

# TODO options window to set this up
_CONFIG_LEGEND = "/gui/WSpectrumCollection/plot_overlapped/flag_legend"


class WSpectrumCollection(a99.WBase):
    """Editor for SpectrumCollection objects"""

    # argument0: flag_changed_header
    edited = pyqtSignal(bool)

    def __init__(self, parent):
        a99.WBase.__init__(self, parent)

        def keep_ref(obj):
            self._refs.append(obj)
            return obj

        # Whether all the values in the fields are valid or not
        self.flag_valid = False
        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False
        self.collection = None # SpectrumCollection


        # # Creates actions

        action = self.action_add_spectra = QAction(a99.get_icon("list-add"), "&Add spectra...", self)
        action.setToolTip("Opens a 'Open File' window where multiple files can be selected.\n"
                          "Accepts any supported spectral type, and also other Spectrum Collection files.\n"
                          "Files are added in alphabetical order.")
        action.triggered.connect(self.on_add_spectra)

        action = self.action_all_group = QAction(a99.get_icon("group-by"), "&Group...", self)
        action.triggered.connect(self.on_all_group)

        # action = self.action_all_use_spectrum_block = QAction(a99.get_icon("go-next"), "&Transform...", self)
        # action.triggered.connect(self.on_all_use_spectrum_block)

        action = self.action_all_to_scalar = QAction(a99.get_icon("go-next"), "To &Scalar...", self)
        action.triggered.connect(self.on_all_to_scalar)

        action = self.action_all_plot_xy = QAction(a99.get_icon("visualization"), "X-&Y Plot", self)
        action.triggered.connect(self.on_plot_xy)

        action = self.action_all_plot_xyz = QAction(a99.get_icon("visualization"), "X-Y-&Z Plot", self)
        action.triggered.connect(self.on_plot_xyz)

        action = self.action_all_export_csv = QAction(a99.get_icon("document-export"), "&Export CSV...", self)
        action.triggered.connect(self.on_all_export_csv)

        action = self.action_sel_use_spectrum_block = QAction(a99.get_icon("go-next"), "&Transform...", self)
        action.setToolTip("Selected spectra will be deleted and transformed spectra will be added at the end")
        action.triggered.connect(self.on_sel_use_spectrum_block)

        action = self.action_sel_plot_stacked = QAction(a99.get_icon("chart2"), "Plot &Stacked", self)
        action.triggered.connect(self.on_sel_plot_stacked)

        action = self.action_sel_plot_overlapped = QAction(a99.get_icon("chart1"), "Plot &Overlapped", self)
        action.triggered.connect(self.on_sel_plot_overlapped)

        action = self.action_sel_open_in_new = QAction(a99.get_icon("window-new"), "Open in new window", self)
        action.triggered.connect(self.on_sel_open_in_new)

        action = self.action_sel_delete = QAction(a99.get_icon("list-remove"), "Delete", self)
        action.triggered.connect(self.on_sel_delete)

        action = self.action_curr_scale = QAction(a99.get_icon("zoom-fit"), "Scale to Magnitude...", self)
        action.triggered.connect(self.on_curr_scale)


        # ## Makes a menu
        menu = self.menu_actions = QMenu("Spectrum &Collection")
        menu.addAction(self.action_add_spectra)
        m = keep_ref(menu.addMenu("With &All Spectra"))
        m.addAction(self.action_all_group)
        # m.addAction(self.action_sel_use_spectrum_block)
        m.addAction(self.action_all_to_scalar)
        m.addAction(self.action_all_plot_xy)
        m.addAction(self.action_all_plot_xyz)
        m.addAction(self.action_all_export_csv)
        m = keep_ref(menu.addMenu("With &Selected Spectra"))
        m.addAction(self.action_sel_use_spectrum_block)
        m.addAction(self.action_sel_plot_stacked)
        m.addAction(self.action_sel_plot_overlapped)
        m.addAction(self.action_sel_open_in_new)
        m.addAction(self.action_sel_delete)
        m = keep_ref(menu.addMenu("With &Current Spectrum"))
        m.addAction(self.action_curr_scale)


        # # Central layout
        # Will have a toolbox and a table. When the toolbox contracts, there is more space for the table
        lwmain = self.centralLayout = QVBoxLayout()
        a99.set_margin(lwmain, 0)
        self.setLayout(lwmain)




        # ## Toolbox
        # The toolbox will have only one widget whose layout is a grid with two columns:
        #   - The first column contains labels
        #   - The second column contains panels of buttons layed horizontally


        wtoolboxes = keep_ref(a99.WCollapsiblePanel())
        lwmain.addWidget(wtoolboxes)
        wtoolboxes.label.setText("<b>Tools</b>")
        wtoolboxes.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        # ### Layout to have (grid layout, spacer)
        lwtoolboxes = QHBoxLayout(wtoolboxes.widget)
        lwmain.addLayout(lwtoolboxes)
        a99.set_margin(lwtoolboxes, 0)
        lwtoolboxes.setSpacing(0)

        # #### Grid layour
        lg = QGridLayout()
        lwtoolboxes.addLayout(lg)
        a99.set_margin(lg, 2)
        lg.setSpacing(0)
        label = keep_ref(QLabel('<b>File:</b>'))
        label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        lg.addWidget(label, 0, 0)
        label = keep_ref(QLabel('<b>With all:</b>'))
        label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        lg.addWidget(label, 1, 0)
        label = keep_ref(QLabel('<b>With selected:</b>'))
        label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        lg.addWidget(label, 2, 0)
        label = keep_ref(QLabel('<b>With current:</b>'))
        label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        lg.addWidget(label, 3, 0)


        TOOL_BUTTON_STYLE = Qt.ToolButtonTextBesideIcon
        # TOOL_BUTTON_STYLE = Qt.ToolButtonIconOnly


        # ##### Toolbar #0: operations independent of the table widget selection state
        ###
        tb = self.toolBar0 = QToolBar()
        lg.addWidget(tb, 0, 1)
        tb.setToolButtonStyle(TOOL_BUTTON_STYLE)
        tb.addAction(self.action_add_spectra)
        tb.setIconSize(QSize(16, 16))

        # ##### Toolbar #1: operations independent of the table widget selection state
        ###
        tb = self.toolBar1 = QToolBar()
        lg.addWidget(tb, 1, 1)
        tb.setToolButtonStyle(TOOL_BUTTON_STYLE)
        tb.setIconSize(QSize(16, 16))
        tb.addAction(self.action_all_group)
        tb.addAction(self.action_all_to_scalar)
        tb.addAction(self.action_all_plot_xy)
        tb.addAction(self.action_all_plot_xyz)
        tb.addAction(self.action_all_export_csv)


        # ##### Toolbar #2: operations affecting only the spectra which are selected
        ###
        tb = self.toolBar2 = QToolBar()
        lg.addWidget(tb, 2, 1)
        tb.setIconSize(QSize(16, 16))
        tb.setToolButtonStyle(TOOL_BUTTON_STYLE)
        tb.addAction(self.action_sel_use_spectrum_block)
        tb.addAction(self.action_sel_plot_stacked)
        tb.addAction(self.action_sel_plot_overlapped)
        tb.addAction(self.action_sel_open_in_new)
        tb.addAction(self.action_sel_delete)

        # ##### Toolbar #3: operations affecting only the current spectrum
        ###

        tb = self.toolBar3 = QToolBar()
        lg.addWidget(tb, 3, 1)
        tb.setIconSize(QSize(16, 16))
        tb.setToolButtonStyle(TOOL_BUTTON_STYLE)
        tb.addAction(self.action_curr_scale)

        # #### Spacer
        lwtoolboxes.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))


        # ## Table Widget
        a = self.twSpectra = QTableWidget()
        lwmain.addWidget(a)
        a.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        a.setSelectionMode(QAbstractItemView.MultiSelection)
        a.setSelectionBehavior(QAbstractItemView.SelectRows)
        a.setAlternatingRowColors(True)
        a.setEditTriggers(QAbstractItemView.DoubleClicked | QAbstractItemView.EditKeyPressed)
        a.setFont(a99.MONO_FONT)
        a.installEventFilter(self)
        a.setContextMenuPolicy(Qt.CustomContextMenu)
        a.customContextMenuRequested.connect(self.on_twSpectra_customContextMenuRequested)
        a.setSortingEnabled(True)
        a.cellChanged.connect(self.on_twSpectra_cellChanged)
        a.itemSelectionChanged.connect(self.on_twSpectra_itemSelectionChanged)
        ah = a.horizontalHeader()
        ah.setSectionsMovable(True)
        ah.setContextMenuPolicy(Qt.CustomContextMenu)
        ah.customContextMenuRequested.connect(self.show_header_context_menu)
        ah.setSelectionMode(QAbstractItemView.SingleSelection)
        ah.sectionMoved.connect(self.section_moved)


        # # Final adjustments
        self.setEnabled(False)  # disabled until load() is called
        a99.style_checkboxes(self)
        self.flag_process_changes = True
        self.add_log("Welcome from %s.__init__()" % (self.__class__.__name__))


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def set_collection(self, x):
        assert isinstance(x, ft.SpectrumCollection)
        self.collection = x
        self.__update_gui()
        self.setEnabled(True)


    def get_selected_spectra(self):
        return [self.collection.spectra[i] for i in self.get_selected_spectrum_indexes()]


    def get_selected_row_indexes(self):
        ii = list(set([index.row() for index in self.twSpectra.selectedIndexes()]))
        return ii


    def get_selected_spectrum_indexes(self):
        items = self.twSpectra.selectedItems()
        ii = []
        for item in items:
            obj = item.data(1)
            if isinstance(obj, int):
                ii.append(obj)
        ii.sort()
        return ii


    def row_index_to_spectrum_index(self, row_index):
        """
        Converts table row index within self.collection.spectra

        This is necessary because the table may be sorted
        """

        item = self.twSpectra.item(row_index, self.twSpectra.columnCount()-1)
        ret = item.data(1)
        return ret


    def get_current_spectrum_index(self):
        row_index = self.twSpectra.currentRow()
        if row_index == -1:
            return -1
        return self.row_index_to_spectrum_index(row_index)


    def get_current_spectrum(self):
        """Returns spectrum on which the table cursor (movable with the keyboar arrows) is currently"""
        spectrum_index = self.get_current_spectrum_index()
        if spectrum_index == -1:
            return None
        return self.collection.spectra[spectrum_index]


    def update(self):
        """Refreshes the GUI to reflect what is in self.collection"""
        self.__update_gui()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        """Sets focus to first field. Note: reason is ignored."""
        self.twSpectra.setFocus()


    def eventFilter(self, source, event):
        if event.type() == QEvent.FocusIn:
            # text = random_name()
            # self.__add_log(text)
            pass

        if event.type() == QEvent.KeyPress:
            if event.key() == Qt.Key_Delete:
                if source == self.twSpectra:
                    n_deleted = self.__delete_spectra()
                    if n_deleted > 0:
                        self.edited.emit(False)
        return False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def section_moved(self, idx_logical, idx_vis_old, idx_vis_new):
        obj = self.collection
        l = self.collection.fieldnames_visible
        fieldname_current = l[idx_vis_old]
        del l[idx_vis_old]
        l.insert(idx_vis_new, fieldname_current)


    def show_header_context_menu(self, position):
        obj = self.collection

        ah = self.twSpectra.horizontalHeader()
        # col_idx = ah.logicalIndexAt(position)
        col_idx = ah.visualIndex(ah.logicalIndexAt(position))

        menu = QMenu()
        act_hide_current = None

        if col_idx < len(obj.fieldnames_visible):
            fieldname_current = obj.fieldnames_visible[col_idx]
            act_hide_current = menu.addAction("&Hide field '%s'" % fieldname_current)
            menu.addSeparator()

        act_show_all = menu.addAction("&Show all fields")
        act_hide_all = menu.addAction("&Hide all fields")
        act_restore_order = menu.addAction("&Restore order")
        menu.addSeparator()

        aa_visible = []
        for fieldname in obj.fieldnames:
            act = menu.addAction(fieldname)
            act.setCheckable(True)
            act.setChecked(fieldname in obj.fieldnames_visible)
            aa_visible.append(act)

        action = menu.exec_(self.twSpectra.mapToGlobal(position))
        flag_update = False
        if action == act_hide_current:
            obj.fieldnames_visible.remove(fieldname_current)
            flag_update = True

        elif action == act_show_all:
            for fieldname in reversed(obj.fieldnames):
                if not fieldname in obj.fieldnames_visible:
                    obj.fieldnames_visible.insert(0, fieldname)
                    flag_update = True

        elif action == act_hide_all:
            obj.fieldnames_visible = []
            flag_update = True

        elif action == act_restore_order:
            curr_visible = copy.copy(obj.fieldnames_visible)
            obj.fieldnames_visible = []
            for fieldname in obj.fieldnames:
                if fieldname in curr_visible:
                    obj.fieldnames_visible.append(fieldname)
            flag_update = not (curr_visible == obj.fieldnames_visible)

        elif action in aa_visible:
            idx = aa_visible.index(action)
            if not aa_visible[idx].isChecked():
                obj.fieldnames_visible.remove(obj.fieldnames[idx])
            else:
                obj.fieldnames_visible.insert(0, obj.fieldnames[idx])
            flag_update = True

        if flag_update:
            self.__update_gui()
            self.edited.emit(False)


    def on_twSpectra_customContextMenuRequested(self, position):
        """Mounts, shows popupmenu for the tableWidget control, and takes action."""
        obj = self.collection
        menu = QMenu()
        act_del = menu.addAction("&Delete selected (Del)")

        action = menu.exec_(self.twSpectra.mapToGlobal(position))
        flag_update = False
        if action == act_del:
            n_deleted = self.__delete_spectra()
            if n_deleted > 0:
                self.edited.emit(False)

        if flag_update:
            self.edited.emit(False)
            self.__update_gui()


    def on_all_group(self):
        # Import here to circumvent cyclic dependency
        from .a_XGroupSpectra import XGroupSpectra
        form = XGroupSpectra()

        if not form.exec_():
            return

        # It sth fails, will restore original
        save = copy.deepcopy(self.collection)
        try:
            grouper = ex.blocks.slb.SLB_UseGroupBlock(form.block, form.group_by)
            self.collection = grouper.use(self.collection)
            self.__update_gui()
            flag_emit = True
        except Exception as E:
            # Restores and logs error
            self.collection = save
            self.add_log_error("Failed to transform spectra: %s" % str(E), True)
            raise

        if flag_emit:
            self.edited.emit(True)

    #
    # def on_all_use_spectrum_block(self):
    #     flag_emit, flag_changed_header = False, False
    #
    #     from .a_XUseSpectrumBlock import XUseSpectrumBlock
    #     form = self.keep_ref(XUseSpectrumBlock())
    #     if not form.exec_():
    #         return
    #
    #     # It sth fails, will restore original
    #     save = copy.deepcopy(self.collection)
    #     try:
    #         sspp = save.spectra
    #         self.collection.clear()
    #         for sp in sspp:
    #             block = copy.deepcopy(form.block)
    #             self.collection.add_spectrum(block.use(sp))
    #
    #         self.__update_gui()
    #         flag_emit = True
    #
    #     except Exception as E:
    #         # Restores and logs error
    #         self.collection = save
    #         self.add_log_error("Failed to transform spectra: %s" % str(E), True)
    #         raise
    #
    #     if flag_emit:
    #         self.edited.emit(True)


    def on_all_to_scalar(self):
        flag_emit, flag_changed_header = False, False

        from .a_XToScalar import XToScalar
        form = self.keep_ref(XToScalar())
        if not form.exec_():
            return

        # It sth fails, will restore original
        save = copy.deepcopy(self.collection)
        try:
            for sp in self.collection.spectra:
                sp.more_headers[form.fieldname] = form.block.use(sp)

            if not form.fieldname in self.collection.fieldnames:
                self.collection.fieldnames.insert(0, form.fieldname)

            if not form.fieldname in self.collection.fieldnames_visible:
                self.collection.fieldnames_visible.insert(0, form.fieldname)

            self.__update_gui()
            flag_emit = True

        except Exception as E:
            # Restores and logs error
            self.add_log_error("Failed to extract scalar: %s" % str(E), True)
            self.collection = save
            raise

        if flag_emit:
            self.edited.emit(True)


    def on_plot_xy(self):
        # Import here to circumvent cyclic dependency
        from .a_XPlotXY import XPlotXY
        form = self.keep_ref(XPlotXY(self.collection))
        form.show()


    def on_plot_xyz(self):
        # Import here to circumvent cyclic dependency
        from .a_XPlotXYZ import XPlotXYZ
        form = self.keep_ref(XPlotXYZ(self.collection))
        form.show()


    def on_all_export_csv(self):
        new_filename = QFileDialog.getSaveFileName(self, "Export text file (CSV format)", "export.csv", "*.csv")[0]
        if new_filename:
            # self.save_dir, _ = os.path.split(str(new_filename))
            try:
                lines = self.collection.to_csv()
                with open(str(new_filename), "w") as file:
                    file.writelines(lines)
            except Exception as E:
                msg = str("Error exporting text file: %s" % a99.str_exc(E))
                self.add_log_error(msg, True)
                raise


    def on_sel_use_spectrum_block(self):
        flag_emit, flag_changed_header = False, False

        from .a_XUseSpectrumBlock import XUseSpectrumBlock
        form = self.keep_ref(XUseSpectrumBlock())
        if not form.exec_():
            return

        # It sth fails, will restore original
        save = copy.deepcopy(self.collection)
        try:
            sspp = self.get_selected_spectra()
            self.collection.delete_spectra(self.get_selected_spectrum_indexes())
            for sp in sspp:
                block = copy.deepcopy(form.block)
                self.collection.add_spectrum(block.use(sp))

            self.__update_gui()
            flag_emit = True

        except Exception as E:
            # Restores and logs error
            self.collection = save
            self.add_log_error("Failed to transform spectra: %s" % str(E), True)
            raise

        if flag_emit:
            self.edited.emit(True)


    def on_sel_plot_stacked(self):
        sspp = self.get_selected_spectra()
        if len(sspp) > 0:
            ex.plot_spectra(sspp)


    def on_sel_plot_overlapped(self):
        sspp = self.get_selected_spectra()
        flag_legend = ex.get_config().get_item(_CONFIG_LEGEND, True)
        oo = ex.PlotSpectrumSetup(flag_legend=flag_legend)
        if len(sspp) > 0:
            ex.plot_spectra_overlapped(sspp, setup=oo)


    def on_sel_open_in_new(self):
        from .a_XFileSpectrumList import XFileSpectrumList
        ii = self.get_selected_spectrum_indexes()
        if len(ii) > 0:
            other = copy.deepcopy(self.collection)
            other.spectra = [other.spectra[i] for i in ii]
            # TODO who said it is a spectrum list? Could well be a FileSparseCube!!!! How to sort this?????????????????????????????????????????????????????????????????????
            f = ft.FileSpectrumList()
            f.splist = other
            f.filename = "noname.dontknowextension"  # TODO figure out a way to determine this filename because I dont know the original filename

            form = self.keep_ref(XFileSpectrumList())
            form.load(f)
            form.show()


    def on_sel_delete(self):
        n = self.__delete_spectra()
        if n > 0:
            self.edited.emit(False)


    def on_curr_scale(self):
        """Performs a scaling operation on the current spectrum"""
        sp = self.get_current_spectrum()

        if sp is None:
            return

        form = XScaleSpectrum()
        form.set_spectrum(sp)
        if form.exec_():
            k = form.factor()
            if k != 1:
                sp.y *= k
                self.__update_gui()
                self.edited.emit(False)


    def on_twSpectra_cellChanged(self, row, column):
        """Cell has been edited manually: commit to self.collection"""
        if self.flag_process_changes:
            flag_emit = False
            text = None
            item = self.twSpectra.item(row, column)
            name = self.__get_tw_header(column)
            self.flag_process_changes = False
            try:
                value = str(item.text())
                # Tries to convert to float, otherwise stores as string
                try:
                    value = float(value)
                except:
                    pass

                # Certain fields must evaluate to integer because they are pixel positions
                if name in ("PIXEL-X", "PIXEL-Y", "Z-START"):
                    value = int(value)

                # Units must be a valid astropy.Unit() argument
                if name in ("X-UNIT", "Y-UNIT"):
                    value = u.Unit(value)

                spectrum_index = self.row_index_to_spectrum_index(row)

                self.collection.spectra[spectrum_index].more_headers[name] = value

                flag_emit = True
                # replaces edited text with eventually cleaner version, e.g. decimals from integers are discarded
                item.setText(str(value))

            except Exception as E:
                # restores original value
                item.setText(str(self.collection.spectra[row].more_headers.get(name)))
                self.add_log_error(a99.str_exc(E), True)
                raise

            finally:
                self.flag_process_changes = True

            if flag_emit:
                self.edited.emit(False)


    def on_twSpectra_itemSelectionChanged(self):
        """Updates actions enabled state"""
        self.__update_enabled_actions()


    def on_add_spectra(self):
        flag_emit = False
        filenames = QFileDialog.getOpenFileNames(self, "Add Spectra", "",
         "All files(*.*);;Spectrum List files (*.splist);;Sparse Cube files (*.sparsecube)")[0]
        if not filenames:
            return

        # classes = ft.classes_sp()+[ft.FileSpectrumList, ft.FileSparseCube]
        classes = ft.classes_collection()
        report, successful, failed = ["<b>Results</b>"], [], []
        for filename in filenames:
            filename = str(filename)
            basename = os.path.basename(filename)
            file = a99.load_with_classes(filename, classes)
            try:
                if file is None:
                    raise RuntimeError("Could not load file")
                if isinstance(file, ft.FileSpectrum):
                    self.collection.add_spectrum(file.spectrum)
                elif isinstance(file, ft.FileSpectrumList):
                    self.collection.merge_with(file.splist)
                elif isinstance(file, ft.FileSparseCube):
                    self.collection.merge_with(file.sparsecube)
                successful.append("  - %s" % basename)
            except Exception as e:
                failed.append('&nbsp;&nbsp;- %s: %s' % (basename, str(e)))
                s = "Error adding file '%s': %s" % (basename, a99.str_exc(e))
                a99.get_python_logger().exception(s)
                self.add_log_error(s)

        if len(successful) > 0:
            report.extend(["", "Successful:"])
            report.extend(successful)

            self.__update_gui()
            flag_emit = True

        if len(failed) > 0:
            report.extend(["", "Failed:"])
            report.extend(failed)

        if flag_emit:
            self.edited.emit(False)

        a99.show_message("<br>".join(report))


    # def on_merge_with(self):
    #     flag_emit = False
    #     try:
    #         # TODO another SpectrumCollection, not SpectrumList
    #         new_filename = QFileDialog.getOpenFileName(self, "Merge with another Spectrum List file", "", "*.splist")[0]
    #         if new_filename:
    #             new_filename = str(new_filename)
    #             f = ft.FileSpectrumList()
    #             f.load(new_filename)
    #             self.collection.merge_with(f.splist)
    #             self.__update_gui()
    #             flag_emit = True
    #     except Exception as E:
    #         msg = "Error merging: %s" % a99.str_exc(E)
    #         self.add_log_error(msg, True)
    #         raise
    #
    #     if flag_emit:
    #         self.edited.emit(False)
    #


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear


    def __get_tw_header(self, column):
        return str(self.twSpectra.horizontalHeaderItem(column).text())


    def __update_gui(self):
        """Updates GUI to match self.collection, i.e., mainly the table widget"""

        self.flag_process_changes = False
        try:
            # Builds table widget contents
            spectra = self.collection.spectra
            t = self.twSpectra
            n = len(spectra)
            FIXED = ["Spectrum summary report"]
            fieldnames_visible = self.collection.fieldnames_visible
            all_headers = fieldnames_visible+FIXED
            nc = len(all_headers)
            a99.reset_table_widget(t, n, nc)
            t.setHorizontalHeaderLabels(all_headers)
            i = 0
            for sp in spectra:
                j = 0

                # Spectrum.more_headers columns
                for h in fieldnames_visible:
                    twi = QTableWidgetItem(str(sp.more_headers.get(h)))
                    if h in "Z-START":  # fields that should be made read-only
                        twi.setFlags(twi.flags() & ~Qt.ItemIsEditable)
                    t.setItem(i, j, twi)
                    j += 1

                # Spectrum self-generated report
                twi = QTableWidgetItem(sp.one_liner_str())
                twi.setFlags(twi.flags() & ~Qt.ItemIsEditable)
                # stores spectrum index not to lose track in case the table is sorted by column
                twi.setData(1, i)
                t.setItem(i, j, twi)
                j += 1

                i += 1

            t.resizeColumnsToContents()

        finally:
            self.__update_enabled_actions()

            self.flag_process_changes = True


    def __update_enabled_actions(self):
        can_group = isinstance(self.collection, ft.SpectrumList)
        has_any = self.twSpectra.rowCount() > 0
        any_selected = len(self.twSpectra.selectedItems()) > 0
        has_current = self.twSpectra.currentRow() > -1

        self.action_add_spectra.setEnabled(True)
        self.action_all_group.setEnabled(can_group and has_any)
        self.action_all_to_scalar.setEnabled(has_any)
        self.action_all_export_csv.setEnabled(has_any)
        self.action_sel_use_spectrum_block.setEnabled(any_selected)
        self.action_sel_plot_stacked.setEnabled(any_selected)
        self.action_sel_plot_overlapped.setEnabled(any_selected)
        self.action_sel_open_in_new.setEnabled(any_selected)
        self.action_sel_delete.setEnabled(any_selected)
        self.action_curr_scale.setEnabled(has_current)

    def __delete_spectra(self):
        ii = self.get_selected_spectrum_indexes()
        if len(ii) > 0:
            self.collection.delete_spectra(ii)
            self.__update_gui()

        return len(ii)



#!/usr/bin/env python3

"""
Prints table of diatomic molecular constants

If formula is specified, prints data for single formula;
otherwise, prints full table
"""

import f311.convmol as cm
import argparse
import logging
import a99


pf.logging_level = logging.INFO

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('-f', type=str, help='Formula of molecule',
                        default='(all)', nargs='?')
    parser.add_argument('-id', type=str, help='State internal ID (second column when printing the whole table)',
                        default=None, nargs='?')
    args = parser.parse_args()

    kwargs = {}
    if not args.f == "(all)":
        kwargs["formula"] = args.f
    if args.id is not None:
        kwargs["state.id"] = args.id

    cm.moldb.print_states(**kwargs)


#!/usr/bin/env python3

"""
Molecular Constants DB editor
"""

import sys
import argparse
import logging
import a99
import logging
import os
import f311.filetypes as ft
import f311.explorer as ex
from ... import convmol as cm


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":

    deffn = ft.FileMolDB.default_filename

    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('fn', type=str, help='SQLite file name',
                        default=deffn, nargs='?')
    args = parser.parse_args()

    if args.fn == deffn and not os.path.isfile(deffn):
        args.fn = None

    m = None
    if args.fn is not None:
        m = ft.FileMolDB()
        m.load(args.fn)
    app = a99.get_QApplication([])
    form = cm.XFileMolDB(None, m)
    form.show()
    sys.exit(app.exec_())
